"""Contains version info dataclass."""
from __future__ import annotations

from dataclasses import dataclass

from pyplumio.version import __version__


def _formated_version(version=__version__) -> str:
    """Format version for program version response."""
    version_parts = version.split(".")
    return ".".join(version_parts[0:3])


@dataclass
class VersionInfo:
    """Represents version info provided in program version response."""

    software: str = _formated_version()
    struct_tag: bytes = b"\xFF\xFF"
    struct_version: int = 5
    device_id: bytes = b"\x7A\x00"
    processor_signature: bytes = b"\x00\x00\x00"



"""Contains frame class."""
from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from enum import IntEnum, unique
from typing import ClassVar, Dict, Final, List, Optional

from pyplumio import util
from pyplumio.const import ADDR_BROADCAST, ADDR_ECONET
from pyplumio.exceptions import UnknownFrameError
from pyplumio.helpers.typing import DeviceDataType, MessageType

FRAME_START: Final = 0x68
FRAME_END: Final = 0x16
HEADER_OFFSET: Final = 0
HEADER_SIZE: Final = 7
FRAME_TYPE_SIZE: Final = 1
CRC_SIZE: Final = 1
DELIMITER_SIZE: Final = 1
ECONET_TYPE: Final = 48
ECONET_VERSION: Final = 5


@unique
class FrameTypes(IntEnum):
    """Contains frame types."""

    REQUEST_STOP_MASTER = 24
    REQUEST_START_MASTER = 25
    REQUEST_CHECK_DEVICE = 48
    REQUEST_BOILER_PARAMETERS = 49
    REQUEST_MIXER_PARAMETERS = 50
    REQUEST_SET_BOILER_PARAMETER = 51
    REQUEST_SET_MIXER_PARAMETER = 52
    REQUEST_SCHEDULES = 54
    REQUEST_SET_SCHEDULE = 55
    REQUEST_UID = 57
    REQUEST_PASSWORD = 58
    REQUEST_BOILER_CONTROL = 59
    REQUEST_ALERTS = 61
    REQUEST_PROGRAM_VERSION = 64
    REQUEST_DATA_SCHEMA = 85
    RESPONSE_DEVICE_AVAILABLE = 176
    RESPONSE_BOILER_PARAMETERS = 177
    RESPONSE_MIXER_PARAMETERS = 178
    RESPONSE_SET_BOILER_PARAMETER = 179
    RESPONSE_SET_MIXER_PARAMETER = 180
    RESPONSE_SCHEDULES = 182
    RESPONSE_UID = 185
    RESPONSE_PASSWORD = 186
    RESPONSE_BOILER_CONTROL = 187
    RESPONSE_ALERTS = 189
    RESPONSE_PROGRAM_VERSION = 192
    RESPONSE_DATA_SCHEMA = 213
    MESSAGE_REGULATOR_DATA = 8
    MESSAGE_SENSOR_DATA = 53


def _handler_class_path(frame_type_name: str) -> str:
    """Return handler class path from module and frame type name."""
    module, type_name = frame_type_name.split("_", 1)

    return f"{module.lower()}s.{util.to_camelcase(type_name)}{module.capitalize()}"


# Dictionary of frame handler classes indexed by frame types.
# example: "24: requests.StopMasterRequest"
FRAME_TYPES: Dict[int, str] = {
    frame_type.value: _handler_class_path(frame_type.name) for frame_type in FrameTypes
}


def is_known_frame_type(frame_type: int) -> bool:
    """Check if frame type is known."""
    try:
        FrameTypes(frame_type)
    except ValueError:
        return False

    return True


def get_frame_handler(frame_type: int) -> str:
    """Return handler class path for the frame type."""
    if frame_type in FRAME_TYPES:
        return f"frames.{FRAME_TYPES[frame_type]}"

    raise UnknownFrameError(f"Unknown frame type ({frame_type})")


@dataclass
class FrameDataClass:
    """Data class mixin for the frame."""

    recipient: int = ADDR_BROADCAST
    sender: int = ADDR_ECONET
    sender_type: int = ECONET_TYPE
    econet_version: int = ECONET_VERSION
    message: MessageType = field(default_factory=bytearray)
    data: DeviceDataType = field(default_factory=dict)


class Frame(ABC, FrameDataClass):
    """Represents base frame class."""

    frame_type: ClassVar[int]

    def __init__(self, *args, **kwargs):
        """Process frame message and data."""
        super().__init__(*args, **kwargs)

        if not self.message:
            # If message not set, create message bytes from data.
            self.message = self.create_message(self.data)

        if self.message and not self.data:
            # If message is set and data is not, decode message.
            self.data = self.decode_message(self.message)

    def __len__(self) -> int:
        """Return frame length."""
        return self.length

    @abstractmethod
    def create_message(self, data: DeviceDataType) -> MessageType:
        """Create frame message."""

    @abstractmethod
    def decode_message(self, message: MessageType) -> DeviceDataType:
        """Decode frame message."""

    @property
    def length(self) -> int:
        """Return frame length."""
        return (
            HEADER_SIZE
            + FRAME_TYPE_SIZE
            + len(self.message)
            + CRC_SIZE
            + DELIMITER_SIZE
        )

    @property
    def header(self) -> bytearray:
        """Return frame header."""
        buffer = bytearray(HEADER_SIZE)
        util.pack_header(
            buffer,
            HEADER_OFFSET,
            FRAME_START,
            self.length,
            self.recipient,
            self.sender,
            self.sender_type,
            self.econet_version,
        )

        return buffer

    @property
    def bytes(self) -> bytes:
        """Return bytes frame representation."""
        data = self.header
        data.append(self.frame_type)
        data += self.message
        data.append(util.crc(data))
        data.append(FRAME_END)
        return bytes(data)

    @property
    def hex(self) -> List[str]:
        """Return hex frame representation."""
        return util.to_hex(self.bytes)


class Request(Frame):
    """Represents request frame."""

    def create_message(self, data: DeviceDataType) -> MessageType:
        """Create frame message."""
        return bytearray()

    def decode_message(self, message: MessageType) -> DeviceDataType:
        """Decode frame message."""
        return {}

    def response(self, **args) -> Optional[Response]:
        """Return response object for current request."""
        return None


class Response(Frame):
    """Represents response frame."""

    def create_message(self, data: DeviceDataType) -> MessageType:
        """Create frame message."""
        return bytearray()

    def decode_message(self, message: MessageType) -> DeviceDataType:
        """Decode frame message."""
        return {}


class Message(Response):
    """Represents message frame."""



"""Contains various helper methods."""
from __future__ import annotations

import functools
import socket
import struct
from typing import List, Optional

from pyplumio.helpers.typing import ParameterDataType

unpack_float = struct.Struct("<f").unpack
unpack_char = struct.Struct("<b").unpack
unpack_short = struct.Struct("<h").unpack
unpack_int = struct.Struct("<i").unpack
unpack_uint = struct.Struct("<I").unpack
unpack_double = struct.Struct("<d").unpack
unpack_int64 = struct.Struct("<q").unpack
unpack_uint64 = struct.Struct("<Q").unpack
pack_header = struct.Struct("<BH4B").pack_into
unpack_header = struct.Struct("<BH4B").unpack_from


def crc(data: bytes) -> int:
    """Return the checksum."""
    return functools.reduce(lambda x, y: x ^ y, data)


def to_hex(data: bytes) -> List[str]:
    """Return a hex representation of byte string."""
    return [f"{data[i]:02X}" for i in range(0, len(data))]


def unpack_ushort(data: bytes) -> int:
    """Unpack a unsigned short number."""
    return int.from_bytes(data, byteorder="little", signed=False)


def unpack_string(data: bytearray, offset: int = 0) -> str:
    """Unpack a string."""
    strlen = data[offset]
    offset += 1
    return data[offset : offset + strlen + 1].decode()


def unpack_parameter(
    data: bytearray, offset: int = 0, size: int = 1
) -> Optional[ParameterDataType]:
    """Unpack a device parameter."""
    if not check_parameter(data[offset : offset + size * 3]):
        return None

    value = unpack_ushort(data[offset : offset + size])
    min_value = unpack_ushort(data[offset + size : offset + 2 * size])
    max_value = unpack_ushort(data[offset + 2 * size : offset + 3 * size])

    return value, min_value, max_value


def check_parameter(data: bytearray) -> bool:
    """Check if parameter contains any bytes besides 0xFF."""
    return any(x for x in data if x != 0xFF)


def ip4_to_bytes(address: str) -> bytes:
    """Convert ip4 address to bytes."""
    return socket.inet_aton(address)


def ip4_from_bytes(data: bytes) -> str:
    """Convert bytes to ip4 address."""
    return socket.inet_ntoa(data)


def ip6_to_bytes(address: str) -> bytes:
    """Convert ip6 address to bytes."""
    return socket.inet_pton(socket.AF_INET6, address)


def ip6_from_bytes(data: bytes) -> str:
    """Convert bytes to ip6 address."""
    return socket.inet_ntop(socket.AF_INET6, data)


def to_camelcase(text: str) -> str:
    """Convert snake_case to CamelCase."""
    replace = {
        "uid": "UID",
        "ecomax": "EcoMAX",
        "ecoster": "EcoSTER",
    }

    return "".join(
        (x.capitalize() or "_") if x.lower() not in replace else replace[x.lower()]
        for x in text.split("_")
    )



"""Contains connection shortcuts and version information."""
from __future__ import annotations

from pyplumio.connection import Connection, SerialConnection, TcpConnection
from pyplumio.helpers.network_info import (
    WLAN_ENCRYPTION_NONE,
    WLAN_ENCRYPTION_UNKNOWN,
    WLAN_ENCRYPTION_WEP,
    WLAN_ENCRYPTION_WPA,
    WLAN_ENCRYPTION_WPA2,
    EthernetParameters,
    WirelessParameters,
)
from pyplumio.version import __version__


def open_serial_connection(*args, **kwargs) -> SerialConnection:
    """Helper function for Serial connection."""
    return SerialConnection(*args, **kwargs)


def open_tcp_connection(*args, **kwargs) -> TcpConnection:
    """Helper function for TCP connection."""
    return TcpConnection(*args, **kwargs)


def ethernet_parameters(**kwargs) -> EthernetParameters:
    """Return instance of ethernet parameters dataclass."""
    return EthernetParameters(status=True, **kwargs)


def wireless_parameters(**kwargs) -> WirelessParameters:
    """Return instance of wireless parameters dataclass."""
    return WirelessParameters(status=True, **kwargs)


__all__ = [
    "Connection",
    "SerialConnection",
    "TcpConnection",
    "open_serial_connection",
    "open_tcp_connection",
    "ethernet_parameters",
    "wireless_parameters",
    "WLAN_ENCRYPTION_NONE",
    "WLAN_ENCRYPTION_UNKNOWN",
    "WLAN_ENCRYPTION_WEP",
    "WLAN_ENCRYPTION_WPA",
    "WLAN_ENCRYPTION_WPA2",
    "__version__",
]



"""Contains reader and writer classes."""
from __future__ import annotations

from asyncio import StreamReader, StreamWriter
from typing import Final, Optional, Tuple

from pyplumio import util
from pyplumio.const import ADDR_BROADCAST, ADDR_ECONET
from pyplumio.exceptions import ChecksumError, ReadError
from pyplumio.frames import FRAME_START, HEADER_SIZE, Frame, get_frame_handler
from pyplumio.helpers.factory import factory
from pyplumio.helpers.timeout import timeout

READER_TIMEOUT: Final = 10
WRITER_TIMEOUT: Final = 10

MIN_FRAME_LENGTH: Final = 10
MAX_FRAME_LENGTH: Final = 1000


class FrameWriter:
    """Represents frame writer."""

    _writer: StreamWriter

    def __init__(self, writer: StreamWriter):
        """Initialize new Frame Writer object."""
        self._writer = writer

    @timeout(WRITER_TIMEOUT)
    async def write(self, frame: Frame) -> None:
        """Write frame to the connection and
        wait for buffer to drain."""
        self._writer.write(frame.bytes)
        await self._writer.drain()

    @timeout(WRITER_TIMEOUT)
    async def close(self) -> None:
        """Close the stream writer."""
        self._writer.close()
        await self._writer.wait_closed()


class FrameReader:
    """Represents frame reader."""

    _reader: StreamReader

    def __init__(self, reader: StreamReader):
        """Initialize new Frame Reader object."""
        self._reader = reader

    async def _read_header(self) -> Tuple[bytes, int, int, int, int, int]:
        """Locate and read frame header."""
        while True:
            header = await self._reader.read(1)
            if header and header[0] == FRAME_START:
                break

        header += await self._reader.read(HEADER_SIZE - 1)
        if len(header) < HEADER_SIZE:
            raise ReadError(f"Header can't be less than {HEADER_SIZE} bytes")

        [
            _,
            length,
            recipient,
            sender,
            sender_type,
            econet_version,
        ] = util.unpack_header(header)

        return header, length, recipient, sender, sender_type, econet_version

    @timeout(READER_TIMEOUT)
    async def read(self) -> Optional[Frame]:
        """Read the frame and return corresponding handler object."""
        (
            header,
            length,
            recipient,
            sender,
            sender_type,
            econet_version,
        ) = await self._read_header()

        if recipient in (ADDR_ECONET, ADDR_BROADCAST):
            # Destination address is econet or broadcast.
            payload = await self._reader.read(length - HEADER_SIZE)
            payload_size = len(payload)

            if (
                length > MAX_FRAME_LENGTH
                or length < MIN_FRAME_LENGTH
                or (payload_size + HEADER_SIZE) != length
            ):
                raise ReadError(f"Unexpected frame size ({payload_size})")

            if payload[-2] != util.crc(header + payload[:-2]):
                raise ChecksumError(f"Incorrect frame checksum ({payload[-2]})")

            return factory(
                get_frame_handler(frame_type=payload[0]),
                recipient=recipient,
                message=payload[1:-2],
                sender=sender,
                sender_type=sender_type,
                econet_version=econet_version,
            )

        return None



"""Contains tasks helper."""
from __future__ import annotations

import asyncio
from typing import Any, Coroutine, Dict, Set


class TaskManager:
    """Helper class for working with asyncio tasks and futures."""

    _tasks: Set[asyncio.Task]
    _events: Dict[str, asyncio.Event]

    def __init__(self):
        self._tasks = set()
        self._events = {}

    def create_task(self, coro: Coroutine[Any, Any, Any]) -> asyncio.Task:
        """Create asyncio Task and store it's reference."""
        task: asyncio.Task = asyncio.create_task(coro)
        self._tasks.add(task)
        task.add_done_callback(self._tasks.discard)
        return task

    def cancel_tasks(self):
        """Cancel all tasks."""
        for task in self._tasks:
            task.cancel()

    async def wait_until_done(self, return_exceptions: bool = True) -> None:
        """Wait for all task to complete."""
        await asyncio.gather(*self._tasks, return_exceptions=return_exceptions)

    def create_event(self, name: str) -> asyncio.Event:
        """Create the event."""
        if name in self.events:
            return self.events[name]

        event = asyncio.Event()
        self._events[name] = event
        return event

    def set_event(self, name: str) -> None:
        """Set the event."""
        if name in self.events:
            event = self.events[name]
            if not event.is_set():
                event.set()

    @property
    def tasks(self) -> Set[asyncio.Task]:
        """Return set of task references."""
        return self._tasks

    @property
    def events(self) -> Dict[str, asyncio.Event]:
        """Return events."""
        return self._events



"""Contains mixer parameter structure decoder."""
from __future__ import annotations

from typing import Dict, Iterable, List, Optional, Tuple

from pyplumio import util
from pyplumio.const import ATTR_MIXER_PARAMETERS
from pyplumio.helpers.typing import DeviceDataType, ParameterDataType
from pyplumio.structures import StructureDecoder, make_device_data

MIXER_PARAMETERS: List[str] = [
    "mix_target_temp",
    "min_mix_target_temp",
    "max_mix_target_temp",
    "low_mix_target_temp",
    "ctrl_weather_mix",
    "mix_heat_curve",
    "parallel_offset_heat_curve",
    "weather_temp_factor",
    "mix_operation",
    "mix_insensitivity",
    "mix_therm_operation",
    "mix_therm_mode",
    "mix_off_therm_pump",
    "mix_summer_work",
]


def _decode_mixer_parameters(
    message: bytearray, offset: int, parameter_name_indexes: Iterable
) -> Tuple[Dict[str, ParameterDataType], int]:
    """Decode parameters for a single mixer."""
    mixer_parameters: Dict[str, ParameterDataType] = {}
    for index in parameter_name_indexes:
        parameter = util.unpack_parameter(message, offset)
        if parameter is not None:
            parameter_name = MIXER_PARAMETERS[index]
            mixer_parameters[parameter_name] = parameter

        offset += 3

    return mixer_parameters, offset


class MixerParametersStructure(StructureDecoder):
    """Represent mixer parameters data structure."""

    def decode(
        self, message: bytearray, offset: int = 0, data: Optional[DeviceDataType] = None
    ) -> Tuple[DeviceDataType, int]:
        """Decode bytes and return message data and offset."""
        first_parameter = message[offset + 1]
        parameters_number = message[offset + 2]
        mixers_number = message[offset + 3]
        offset += 4
        mixer_parameters: List[Dict[str, ParameterDataType]] = []
        for _ in range(mixers_number):
            parameters, offset = _decode_mixer_parameters(
                message,
                offset,
                range(first_parameter, parameters_number + first_parameter),
            )
            mixer_parameters.append(parameters)

        return (
            make_device_data(data, {ATTR_MIXER_PARAMETERS: mixer_parameters}),
            offset,
        )



"""Contains connection representation."""
from __future__ import annotations

from abc import ABC, abstractmethod
import asyncio
from collections.abc import MutableMapping
import logging
from typing import Any, Final, Tuple

from serial import SerialException
import serial_asyncio

from pyplumio.exceptions import ConnectionFailedError
from pyplumio.helpers.timeout import timeout
from pyplumio.protocol import Protocol

_LOGGER = logging.getLogger(__name__)

CONNECT_TIMEOUT: Final = 5
RECONNECT_TIMEOUT: Final = 20


class Connection(ABC):
    """Base connection representation. All specific connection classes
    must me inherited from this class."""

    _kwargs: MutableMapping[str, Any]
    _protocol: Protocol
    _closing: bool = False
    _reconnect_on_failure: bool = True

    def __init__(
        self,
        reconnect_on_failure: bool = True,
        ethernet_parameters=None,
        wireless_parameters=None,
        **kwargs,
    ):
        """Initialize Connection object."""
        self._kwargs = kwargs
        self._closing = False
        self._protocol = Protocol(
            self._connection_lost_callback,
            ethernet_parameters,
            wireless_parameters,
        )
        self._reconnect_on_failure = reconnect_on_failure

    async def __aenter__(self):
        """Provide entry point for context manager."""
        await self.connect()
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        """Provide exit point for context manager."""
        await self.close()

    def __getattr__(self, name: str):
        """Return attributes from the underlying protocol object."""
        return getattr(self.protocol, name)

    async def _connection_lost_callback(self) -> None:
        """Callback to resume the connection on connection lost."""
        if self._reconnect_on_failure and not self._closing:
            _LOGGER.error("Connection to the device lost. Trying to reconnect")
            await self._reconnect()

    async def _connect(self) -> None:
        """Establish connection and initialize the protocol object."""
        try:
            reader, writer = await self._open_connection()
            self._protocol.connection_established(reader, writer)
        except (
            OSError,
            SerialException,
            asyncio.TimeoutError,
        ) as connection_error:
            raise ConnectionFailedError from connection_error

    async def _reconnect(self) -> None:
        """Establish connection and reconnect on failure."""
        while True:
            try:
                await self._connect()
                return
            except ConnectionFailedError:
                _LOGGER.error(
                    "Can't connect to the device, retrying in %.1f seconds",
                    RECONNECT_TIMEOUT,
                )
                await asyncio.sleep(RECONNECT_TIMEOUT)

    async def connect(self) -> None:
        """Initialize the connection. This method initializes
        connection via connect or reconnect routines, depending
        on _reconnect_on_failure property."""
        if self._reconnect_on_failure:
            await self._reconnect()
        else:
            await self._connect()

    async def close(self) -> None:
        """Close the connection."""
        self._closing = True
        await self.protocol.shutdown()

    @property
    def protocol(self) -> Protocol:
        """Return protocol object."""
        return self._protocol

    @abstractmethod
    async def _open_connection(
        self,
    ) -> Tuple[asyncio.StreamReader, asyncio.StreamWriter]:
        """Open the connection and return reader and writer objects."""


class TcpConnection(Connection):
    """Represents TCP connection."""

    def __init__(self, host: str, port: int, **kwargs):
        """Initialize TCP connection object."""
        super().__init__(**kwargs)
        self.host = host
        self.port = port

    def __repr__(self):
        """Return string representation of the class."""
        return (
            f"TcpConnection(host={self.host}, port={self.port}, kwargs={self._kwargs})"
        )

    @timeout(CONNECT_TIMEOUT)
    async def _open_connection(
        self,
    ) -> Tuple[asyncio.StreamReader, asyncio.StreamWriter]:
        """Open the connection and return reader and writer objects."""
        return await asyncio.open_connection(
            host=self.host, port=self.port, **self._kwargs
        )


class SerialConnection(Connection):
    """Represents Serial connection."""

    def __init__(self, device: str, baudrate: int = 115200, **kwargs):
        """Initialize Serial connection object."""
        super().__init__(**kwargs)
        self.device = device
        self.baudrate = baudrate

    def __repr__(self):
        """Return string representation of the class."""
        return (
            f"SerialConnection(device={self.device}, baudrate={self.baudrate}, "
            + f"kwargs={self._kwargs})"
        )

    @timeout(CONNECT_TIMEOUT)
    async def _open_connection(
        self,
    ) -> Tuple[asyncio.StreamReader, asyncio.StreamWriter]:
        """Open the connection and return reader and writer objects."""
        return await serial_asyncio.open_serial_connection(
            url=self.device,
            baudrate=self.baudrate,
            bytesize=serial_asyncio.serial.EIGHTBITS,
            parity=serial_asyncio.serial.PARITY_NONE,
            stopbits=serial_asyncio.serial.STOPBITS_ONE,
            **self._kwargs,
        )



"""Contains alarms structure decoder."""
from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from typing import Dict, List, Optional, Tuple

from pyplumio import util
from pyplumio.const import ATTR_ALERTS
from pyplumio.helpers.typing import DeviceDataType
from pyplumio.structures import StructureDecoder, make_device_data


def _convert_to_datetime(seconds: int) -> datetime:
    """Converts timestamp to datetime."""
    intervals = (
        ("year", 32140800),  # 60 * 60 * 24 * 31 * 12
        ("month", 2678400),  # 60 * 60 * 24 * 31
        ("day", 86400),  # 60 * 60 * 24
        ("hour", 3600),  # 60 * 60
        ("minute", 60),
        ("second", 1),
    )

    result: Dict[str, int] = {}

    for name, count in intervals:
        value = seconds // count
        seconds -= value * count
        result[name] = value

    return datetime(
        year=(result["year"] + 2000),
        month=(result["month"] + 1),
        day=(result["day"] + 1),
        hour=result["hour"],
        minute=result["minute"],
        second=result["second"],
    )


@dataclass
class Alert:
    """Represents device alert."""

    code: int
    from_dt: datetime
    to_dt: Optional[datetime]


class AlertsStructure(StructureDecoder):
    """Represents alerts data structure."""

    def decode(
        self, message: bytearray, offset: int = 0, data: Optional[DeviceDataType] = None
    ) -> Tuple[DeviceDataType, int]:
        """Decode bytes and return message data and offset."""
        first_alert = message[offset + 1]
        alerts_number = message[offset + 2]
        offset += 3
        alerts: List[Alert] = []
        for _ in range(first_alert, first_alert + alerts_number):
            code = message[offset]
            from_ts = util.unpack_uint(message[offset + 1 : offset + 5])[0]
            from_dt = _convert_to_datetime(from_ts)
            to_dt = None
            if util.check_parameter(message[offset + 5 : offset + 9]):
                to_ts = util.unpack_uint(message[offset + 5 : offset + 9])[0]
                to_dt = _convert_to_datetime(to_ts)

            alerts.append(Alert(code, from_dt, to_dt))
            offset += 9

        return make_device_data(data, {ATTR_ALERTS: alerts}), offset



"""Contains temperatures structure decoder."""
from __future__ import annotations

import math
from typing import Final, Optional, Tuple

from pyplumio import util
from pyplumio.helpers.typing import DeviceDataType
from pyplumio.structures import StructureDecoder, make_device_data

HEATING_TEMP: Final = "heating_temp"
FEEDER_TEMP: Final = "feeder_temp"
WATER_HEATER_TEMP: Final = "water_heater_temp"
OUTSIDE_TEMP: Final = "outside_temp"
RETURN_TEMP: Final = "return_temp"
EXHAUST_TEMP: Final = "exhaust_temp"
OPTICAL_TEMP: Final = "optical_temp"
UPPER_BUFFER_TEMP: Final = "upper_buffer_temp"
LOWER_BUFFER_TEMP: Final = "lower_buffer_temp"
UPPER_SOLAR_TEMP: Final = "upper_solar_temp"
LOWER_SOLAR_TEMP: Final = "lower_solar_temp"
FIREPLACE_TEMP: Final = "fireplace_temp"
TOTAL_GAIN: Final = "total_gain"
HYDRAULIC_COUPLER_TEMP: Final = "hydraulic_coupler_temp"
EXCHANGER_TEMP: Final = "exchanger_temp"
AIR_IN_TEMP: Final = "air_in_temp"
AIR_OUT_TEMP: Final = "air_out_temp"
TEMPERATURES: Tuple[str, ...] = (
    HEATING_TEMP,
    FEEDER_TEMP,
    WATER_HEATER_TEMP,
    OUTSIDE_TEMP,
    RETURN_TEMP,
    EXHAUST_TEMP,
    OPTICAL_TEMP,
    UPPER_BUFFER_TEMP,
    LOWER_BUFFER_TEMP,
    UPPER_SOLAR_TEMP,
    LOWER_SOLAR_TEMP,
    FIREPLACE_TEMP,
    TOTAL_GAIN,
    HYDRAULIC_COUPLER_TEMP,
    EXCHANGER_TEMP,
    AIR_IN_TEMP,
    AIR_OUT_TEMP,
)


class TemperaturesStructure(StructureDecoder):
    """Represents temperatures data structures."""

    def decode(
        self, message: bytearray, offset: int = 0, data: Optional[DeviceDataType] = None
    ) -> Tuple[DeviceDataType, int]:
        """Decode bytes and return message data and offset."""
        temp_number = message[offset]
        offset += 1
        data = make_device_data(data)
        for _ in range(temp_number):
            index = message[offset]
            temp = util.unpack_float(message[offset + 1 : offset + 5])[0]

            if (not math.isnan(temp)) and 0 <= index < len(TEMPERATURES):
                # Temperature exists and index is in the correct range.
                data[TEMPERATURES[index]] = temp

            offset += 5

        return data, offset



"""Contains schedule helpers."""


from collections.abc import Iterable, MutableMapping
from dataclasses import dataclass
import datetime as dt
import math
from typing import TYPE_CHECKING, Final, Iterator, List, Literal, Tuple

from pyplumio.const import (
    ATTR_PARAMETER,
    ATTR_SCHEDULE,
    ATTR_SCHEDULES,
    ATTR_SWITCH,
    ATTR_TYPE,
    STATE_OFF,
    STATE_ON,
)
from pyplumio.helpers.factory import factory
from pyplumio.helpers.typing import DeviceDataType

if TYPE_CHECKING:
    from pyplumio.devices import Device
else:
    Device = object

TIME_FORMAT: Final = "%H:%M"
START_OF_DAY: Final = "00:00"
END_OF_DAY: Final = "00:00"

STATE_NIGHT: Final = STATE_OFF
STATE_DAY: Final = STATE_ON


def _parse_interval(start: str, end: str) -> Tuple[int, int]:
    """Parse interval string."""
    start_dt = dt.datetime.strptime(start, TIME_FORMAT)
    end_dt = dt.datetime.strptime(end, TIME_FORMAT)
    start_of_day_dt = dt.datetime.strptime(START_OF_DAY, TIME_FORMAT)
    if end_dt == start_of_day_dt:
        # Upper bound of interval is midnight.
        end_dt += dt.timedelta(days=1)

    if end_dt <= start_dt:
        raise ValueError(
            f"Invalid interval ({start}, {end}). Lower boundary must be less than upper."
        )

    start_index = math.floor((start_dt - start_of_day_dt).total_seconds() // (60 * 30))
    stop_index = math.floor((end_dt - start_of_day_dt).total_seconds() // (60 * 30))

    return start_index, stop_index


def _collect_schedule_data(name: str, device: Device) -> DeviceDataType:
    """Return schedule data collected from the device."""
    switch_key = f"schedule_{name}_{ATTR_SWITCH}"
    parameter_key = f"schedule_{name}_{ATTR_PARAMETER}"

    return {
        ATTR_TYPE: name,
        ATTR_SWITCH: device.data[switch_key],
        ATTR_PARAMETER: device.data[parameter_key],
        ATTR_SCHEDULE: device.data[ATTR_SCHEDULES][name],
    }


class ScheduleDay(MutableMapping):
    """Represents single day of schedule."""

    _intervals: List[bool]

    def __init__(self, intervals: List[bool]):
        """Initialize new schedule day object."""
        self._intervals = intervals

    def __repr__(self) -> str:
        """Return serializable representation of the class."""
        return f"ScheduleDay({self._intervals})"

    def __len__(self) -> int:
        """Return schedule length."""
        return len(self._intervals)

    def __iter__(self) -> Iterator[bool]:
        """Return iterator."""
        return self._intervals.__iter__()

    def __getitem__(self, index):
        """Return item."""
        return self._intervals.__getitem__(index)

    def __delitem__(self, index) -> None:
        return self._intervals.__delitem__(index)

    def __setitem__(self, index, value) -> None:
        return self._intervals.__setitem__(index, value)

    def append(self, item) -> None:
        """Append value to interval."""
        self._intervals.append(item)

    def set_state(
        self,
        state: Literal["off", "on"],
        start: str = START_OF_DAY,
        end: str = END_OF_DAY,
    ) -> None:
        """Set state for interval."""
        index, stop_index = _parse_interval(start, end)
        while index < stop_index:
            self._intervals[index] = state == STATE_ON
            index += 1

    def set_on(self, start: str = START_OF_DAY, end: str = END_OF_DAY) -> None:
        """Set on state for interval."""
        self.set_state(STATE_ON, start, end)

    def set_off(self, start: str = START_OF_DAY, end: str = END_OF_DAY) -> None:
        """Set off state for interval."""
        self.set_state(STATE_OFF, start, end)

    @property
    def intervals(self) -> List[bool]:
        """Return intervals."""
        return self._intervals


@dataclass
class Schedule(Iterable):
    """Represents weekly schedule."""

    name: str
    device: Device
    monday: ScheduleDay
    tuesday: ScheduleDay
    wednesday: ScheduleDay
    thursday: ScheduleDay
    friday: ScheduleDay
    saturday: ScheduleDay
    sunday: ScheduleDay

    def __iter__(self) -> Iterator[ScheduleDay]:
        """Return list of days."""
        return (
            self.sunday,
            self.monday,
            self.tuesday,
            self.wednesday,
            self.thursday,
            self.friday,
            self.saturday,
        ).__iter__()

    def commit(self) -> None:
        """Commit changes to the device."""
        self.device.queue.put_nowait(
            factory(
                "frames.requests.SetScheduleRequest",
                recipient=self.device.address,
                data=_collect_schedule_data(self.name, self.device),
            )
        )



"""Contains structure decoders."""
from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Optional, Tuple

from pyplumio.frames import Frame
from pyplumio.helpers.typing import DeviceDataType


def make_device_data(data: Optional[DeviceDataType], *args) -> DeviceDataType:
    """Make new or merge multiple device data."""
    if data is None:
        data = {}

    for new_data in args:
        data = dict(data, **new_data)

    return data


@dataclass
class StructureDataClass:
    "Represents structure dataclass mixin."

    frame: Frame


class Structure(ABC, StructureDataClass):
    """Represents data structure."""

    @abstractmethod
    def encode(self, data: DeviceDataType) -> bytearray:
        """Encode device data to bytearray message."""

    @abstractmethod
    def decode(
        self,
        message: bytearray,
        offset: int = 0,
        data: Optional[DeviceDataType] = None,
    ) -> Tuple[DeviceDataType, int]:
        """Decode bytes and return message data and offset."""


class StructureDecoder(Structure, ABC):
    """Represent structure that only handles decoding."""

    def encode(self, data: DeviceDataType) -> bytearray:
        """Encode device data to bytearray message."""
        return bytearray()

    @abstractmethod
    def decode(
        self, message: bytearray, offset: int = 0, data: Optional[DeviceDataType] = None
    ) -> Tuple[DeviceDataType, int]:
        """Decode bytes and return message data and offset."""



"""Contains device representations."""
from __future__ import annotations

from abc import ABC
import asyncio
from enum import IntEnum, unique
from typing import ClassVar, Dict, List, Optional
from warnings import warn

from pyplumio import util
from pyplumio.const import ATTR_FRAME_VERSIONS
from pyplumio.exceptions import ParameterNotFoundError, UnknownDeviceError
from pyplumio.frames import Frame
from pyplumio.helpers.frame_versions import DEFAULT_FRAME_VERSION, FrameVersions
from pyplumio.helpers.parameter import Parameter
from pyplumio.helpers.task_manager import TaskManager
from pyplumio.helpers.typing import (
    DeviceDataType,
    NumericType,
    SensorCallbackType,
    VersionsInfoType,
)


@unique
class DeviceTypes(IntEnum):
    """Contains device types."""

    ECOMAX = 69
    ECOSTER = 81


def _handler_class_path(device_type_name: str) -> str:
    """Return handler class path from device type name."""
    return f"{device_type_name.lower()}.{util.to_camelcase(device_type_name)}"


# Dictionary of device handler classes indexed by device types.
# example: "69: ecomax.EcoMAX"
DEVICE_TYPES: Dict[int, str] = {
    device_type.value: _handler_class_path(device_type.name)
    for device_type in DeviceTypes
}


def get_device_handler(device_type: int) -> str:
    """Return module and class for device type."""
    if device_type in DEVICE_TYPES:
        return "devices." + DEVICE_TYPES[device_type]

    raise UnknownDeviceError(f"Unknown device ({device_type})")


class AsyncDevice(ABC, TaskManager):
    """Represents a device with awaitable properties."""

    data: DeviceDataType
    _callbacks: Dict[str, List[SensorCallbackType]]

    def __init__(self):
        """Initializes Async Device object."""
        super().__init__()
        self.data = {}
        self._callbacks = {}

    async def get_value(self, name: str, timeout: Optional[float] = None):
        """Return a value. When encountering a parameter, only it's
        value will be returned. To return the Parameter object use
        get_parameter method."""
        if name not in self.data:
            await asyncio.wait_for(self.create_event(name).wait(), timeout=timeout)

        value = self.data[name]
        return int(value) if isinstance(value, Parameter) else value

    async def set_value(
        self, name: str, value: NumericType, timeout: Optional[float] = None
    ) -> None:
        """Set parameter value. Name should point
        to a valid parameter object."""
        if name not in self.data:
            await asyncio.wait_for(self.create_event(name).wait(), timeout=timeout)

        parameter = self.data[name]
        if isinstance(parameter, Parameter):
            parameter.set(value)
            return

        raise ParameterNotFoundError(f"Parameter not found ({name})")

    async def get_parameter(
        self, name: str, timeout: Optional[float] = None
    ) -> Parameter:
        """Return a parameter."""
        if name not in self.data:
            await asyncio.wait_for(self.create_event(name).wait(), timeout=timeout)

        parameter = self.data[name]
        if isinstance(parameter, Parameter):
            return parameter

        raise ParameterNotFoundError(f"Parameter not found ({name})")

    def set_device_data(self, *args, **kwargs) -> None:
        """Call registered callbacks on value change."""
        self.create_task(self.async_set_device_data(*args, **kwargs))

    async def async_set_device_data(self, name: str, value) -> None:
        """Asynchronously call registered callbacks on value change."""
        if name in self._callbacks:
            callbacks = self._callbacks[name].copy()
            for callback in callbacks:
                return_value = await callback(value)
                value = return_value if return_value is not None else value

        self.data[name] = value
        self.set_event(name)

    async def shutdown(self) -> None:
        """Cancel scheduled tasks."""
        self.cancel_tasks()
        await self.wait_until_done()

    def register_callback(self, name: str, callback: SensorCallbackType) -> None:
        """Register a callback for a value change."""
        warn(
            "Function 'register_callback' is deprecated. Use 'subscribe' instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        self.subscribe(name, callback)

    def remove_callback(self, name: str, callback: SensorCallbackType) -> None:
        """Remove the callback."""
        warn(
            "Function 'remove_callback' is deprecated. Use 'unsubscribe' instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        self.unsubscribe(name, callback)

    def subscribe(self, name: str, callback: SensorCallbackType) -> None:
        """Register a callback for a value change."""
        if name not in self._callbacks:
            self._callbacks[name] = []

        self._callbacks[name].append(callback)

    def subscribe_once(self, name: str, callback: SensorCallbackType) -> None:
        """Register a callback for a single call."""

        async def _callback(value):
            """Unregister the callback once it was called."""
            try:
                return await callback(value)
            finally:
                self.unsubscribe(name, _callback)

        self.subscribe(name, _callback)

    def unsubscribe(self, name: str, callback: SensorCallbackType) -> None:
        """Remove the callback."""
        if name in self._callbacks and callback in self._callbacks[name]:
            self._callbacks[name].remove(callback)


class Device(AsyncDevice):
    """Represents base device."""

    address: ClassVar[int]
    queue: asyncio.Queue
    _required_frames: List[int] = []

    def __init__(self, queue: asyncio.Queue):
        """Initialize new Device object."""
        super().__init__()
        self.queue = queue
        versions = FrameVersions(device=self)
        self.subscribe_once(ATTR_FRAME_VERSIONS, self._merge_required_frames)
        self.subscribe(ATTR_FRAME_VERSIONS, versions.async_update)

    async def _merge_required_frames(
        self, frame_versions: VersionsInfoType
    ) -> VersionsInfoType:
        """Merge required frames into version list."""
        requirements = {int(x): DEFAULT_FRAME_VERSION for x in self.required_frames}
        return {**requirements, **frame_versions}

    def handle_frame(self, frame: Frame) -> None:
        """Handle received frame."""
        if frame.data is not None:
            for name, value in frame.data.items():
                self.set_device_data(name, value)

    @property
    def required_frames(self) -> List[int]:
        """Return list of required frame types."""
        return self._required_frames


class Mixer(AsyncDevice):
    """Represents mixer device."""

    def __init__(self, index: int = 0):
        """Initialize new Mixer object."""
        super().__init__()
        self.index = index



"""Contains output flags structure decoder."""
from __future__ import annotations

from typing import Final, Optional, Tuple

from pyplumio import util
from pyplumio.helpers.typing import DeviceDataType
from pyplumio.structures import StructureDecoder, make_device_data

HEATING_PUMP_FLAG: Final = "heating_pump_flag"
WATER_HEATER_PUMP_FLAG: Final = "water_heater_pump_flag"
CIRCULATION_PUMP_FLAG: Final = "circulation_pump_flag"
SOLAR_PUMP_FLAG: Final = "solar_pump_flag"
OUTPUT_FLAGS: Tuple[str, ...] = (
    HEATING_PUMP_FLAG,
    WATER_HEATER_PUMP_FLAG,
    CIRCULATION_PUMP_FLAG,
    SOLAR_PUMP_FLAG,
)


class OutputFlagsStructure(StructureDecoder):
    """Represents output flags structure."""

    def decode(
        self, message: bytearray, offset: int = 0, data: Optional[DeviceDataType] = None
    ) -> Tuple[DeviceDataType, int]:
        """Decode bytes and return message data and offset."""
        output_flags = util.unpack_ushort(message[offset : offset + 4])
        data = make_device_data(
            data,
            {
                HEATING_PUMP_FLAG: bool(output_flags & 0x04),
                WATER_HEATER_PUMP_FLAG: bool(output_flags & 0x08),
                CIRCULATION_PUMP_FLAG: bool(output_flags & 0x10),
                SOLAR_PUMP_FLAG: bool(output_flags & 0x800),
            },
        )
        return data, offset + 4



"""Contains exceptions."""
from __future__ import annotations


class PyPlumIOError(Exception):
    """Base PyPlumIO error class."""


class ConnectionFailedError(PyPlumIOError):
    """Raised on connection failure."""


class ParameterNotFoundError(PyPlumIOError):
    """Raised when device parameter is not found."""


class UnknownDeviceError(PyPlumIOError):
    """Raised on unsupported device."""


class ReadError(PyPlumIOError):
    """Raised on read error."""


class FrameError(PyPlumIOError):
    """Base class for frame errors."""


class ChecksumError(FrameError):
    """Raised on checksum error while parsing frame content."""


class VersionError(FrameError):
    """Raised on unknown frame version."""


class UnknownFrameError(FrameError):
    """Raised on unknown frame type."""


class FrameDataError(FrameError):
    """Raised on incorrect frame data."""



"""Contains factory helper."""
from __future__ import annotations

import importlib
import logging

_LOGGER = logging.getLogger(__name__)


def factory(class_path: str, **kwargs):
    """Return object from class path."""
    try:
        module_name, class_name = class_path.rsplit(".", 1)
        cls = getattr(
            importlib.import_module("." + module_name, "pyplumio"), class_name
        )
        return cls(**kwargs)
    except Exception:
        _LOGGER.error("Failed to load module (%s)", class_path)
        raise



"""Contains schedule decoder."""

from typing import Final, List, Optional, Sequence, Tuple, Union

from pyplumio.const import (
    ATTR_PARAMETER,
    ATTR_SCHEDULE,
    ATTR_SCHEDULES,
    ATTR_SWITCH,
    ATTR_TYPE,
)
from pyplumio.exceptions import FrameDataError
from pyplumio.helpers.typing import DeviceDataType
from pyplumio.structures import Structure, make_device_data
from pyplumio.util import unpack_parameter

SCHEDULE_SIZE: Final = 42  # 6 bytes per day, 7 days total

SCHEDULES: Tuple[str, ...] = (
    "heating",
    "water_heater",
    "circulation_pump",
    "boiler_work",
    "boiler_clean",
    "hear_exchanger_clean",
    "mixer_1",
    "mixer_2",
    "mixer_3",
    "mixer_4",
    "mixer_5",
    "mixer_6",
    "mixer_7",
    "mixer_8",
    "mixer_9",
    "mixer_10",
    "thermostat_1",
    "thermostat_2",
    "thermostat_3",
    "circuit_1",
    "circuit_2",
    "circuit_3",
    "circuit_4",
    "circuit_5",
    "circuit_6",
    "circuit_7",
    "panel_1",
    "panel_2",
    "panel_3",
    "panel_4",
    "panel_5",
    "panel_6",
    "panel_7",
    "main_heater_solar",
    "heating_circulation",
    "internal_thermostat",
    "heater",
    "water_heater_2",
    "intake",
    "intake_summer",
)


def _split_byte(byte: int) -> List[bool]:
    """Split single byte into an eight bits."""
    bits = []
    for bit in reversed(range(8)):
        bits.append(bool(byte & (1 << bit)))

    return bits


def _join_byte(bits: Sequence[Union[int, bool]]) -> int:
    """Join eight bits into a single byte."""
    result = 0
    for bit in bits:
        result = (result << 1) | bit

    return result


def _decode_schedule(message: bytearray, offset: int) -> Tuple[List[List[bool]], int]:
    """Return schedule data and offset."""
    schedule: List[List[bool]] = []
    day: List[bool] = []

    last_offset = offset + SCHEDULE_SIZE
    while offset < last_offset:
        if len(day) == 48:
            schedule.append(day)
            day = []

        day.extend(_split_byte(message[offset]))
        offset += 1

    schedule.append(day)

    return schedule, offset


class SchedulesStructure(Structure):
    """Represents schedule data structure."""

    def encode(self, data: DeviceDataType) -> bytearray:
        """Encode device data to bytearray message."""
        message = bytearray()
        message.append(1)
        try:
            message.append(SCHEDULES.index(data[ATTR_TYPE]))
            message.append(int(data[ATTR_SWITCH]))
            message.append(int(data[ATTR_PARAMETER]))
            schedule = data[ATTR_SCHEDULE]
        except (KeyError, ValueError) as e:
            raise FrameDataError from e

        schedule_bytes = []
        for day in list(schedule):
            schedule_bytes += [
                _join_byte(day[i : i + 8]) for i in range(0, len(day), 8)
            ]

        message += bytearray(schedule_bytes)

        return message

    def decode(
        self, message: bytearray, offset: int = 0, data: Optional[DeviceDataType] = None
    ) -> Tuple[DeviceDataType, int]:
        """Decode bytes and return message data and offset."""
        first_block = message[offset + 1]
        block_number = message[offset + 2]
        offset += 3
        schedules = {}
        for _ in range(first_block, first_block + block_number):
            schedule_type = message[offset]
            schedule_name = SCHEDULES[schedule_type]
            schedule: DeviceDataType = {}
            schedule[ATTR_SWITCH] = (message[offset + 1], 0, 1)
            schedule[ATTR_PARAMETER] = unpack_parameter(message, offset + 2)
            schedule[ATTR_SCHEDULE], offset = _decode_schedule(message, offset + 5)
            schedules[schedule_name] = schedule

        return make_device_data(data, {ATTR_SCHEDULES: schedules}), offset



"""Contains product info dataclasses."""
from __future__ import annotations

from dataclasses import dataclass
from typing import Optional


@dataclass
class ProductInfo:
    """Represents product info provided by UID response."""

    type: int = 0
    product: int = 0
    uid: Optional[str] = None
    logo: int = 0
    image: int = 0
    model: Optional[str] = None


@dataclass
class ConnectedModules:
    """Represents firmware version info."""

    module_a: Optional[str] = None
    module_b: Optional[str] = None
    module_c: Optional[str] = None
    module_lambda: Optional[str] = None
    module_ecoster: Optional[str] = None
    module_panel: Optional[str] = None



"""Contains thermostats structure decoder."""
from __future__ import annotations

from typing import Final, List, Optional, Tuple

from pyplumio import util
from pyplumio.const import ATTR_THERMOSTATS
from pyplumio.helpers.typing import DeviceDataType
from pyplumio.structures import StructureDecoder, make_device_data

ECOSTER_CONTACTS: Final = "contacts"
ECOSTER_SCHEDULE: Final = "schedule"
ECOSTER_MODE: Final = "mode"
ECOSTER_TEMP: Final = "temp"
ECOSTER_TARGET: Final = "target"


class ThermostatsStructure(StructureDecoder):
    """Represents thermostats data structure."""

    def decode(
        self, message: bytearray, offset: int = 0, data: Optional[DeviceDataType] = None
    ) -> Tuple[DeviceDataType, int]:
        """Decode bytes and return message data and offset."""
        if message[offset] == 0xFF:
            return make_device_data(data), offset + 1

        therm_contacts = message[offset]
        offset += 1
        therm_number = message[offset]
        offset += 1
        thermostats: List[DeviceDataType] = []
        contact_mask = 1
        schedule_mask = 1 << 3
        for _ in range(therm_number):
            therm: DeviceDataType = {}
            therm[ECOSTER_CONTACTS] = bool(therm_contacts & contact_mask)
            therm[ECOSTER_SCHEDULE] = bool(therm_contacts & schedule_mask)
            therm[ECOSTER_MODE] = message[offset]
            therm[ECOSTER_TEMP] = util.unpack_float(message[offset + 1 : offset + 5])[0]
            therm[ECOSTER_TARGET] = util.unpack_float(message[offset + 5 : offset + 9])[
                0
            ]
            thermostats.append(therm)
            offset += 9
            contact_mask = contact_mask << 1
            schedule_mask = schedule_mask << 1

        return make_device_data(data, {ATTR_THERMOSTATS: thermostats}), offset



"""Contains frame versions structure decoder."""
from __future__ import annotations

from typing import Optional, Tuple

from pyplumio import util
from pyplumio.const import ATTR_FRAME_VERSIONS
from pyplumio.helpers.typing import DeviceDataType, VersionsInfoType
from pyplumio.structures import StructureDecoder, make_device_data


class FrameVersionsStructure(StructureDecoder):
    """Represents frame version data structure."""

    def decode(
        self, message: bytearray, offset: int = 0, data: Optional[DeviceDataType] = None
    ) -> Tuple[DeviceDataType, int]:
        """Decode bytes and return message data and offset."""
        frame_versions: VersionsInfoType = {}
        frames_number = message[offset]
        offset += 1
        for _ in range(frames_number):
            frame_type = message[offset]
            version = util.unpack_ushort(message[offset + 1 : offset + 3])
            frame_versions[frame_type] = version
            offset += 3

        return make_device_data(data, {ATTR_FRAME_VERSIONS: frame_versions}), offset



"""Contains ecoSTER device representation."""
from __future__ import annotations

from typing import ClassVar

from pyplumio.devices import Device, DeviceTypes


class EcoSTER(Device):
    """Represents ecoSTER thermostat."""

    address: ClassVar[int] = DeviceTypes.ECOSTER



"""Contains device parameter representation."""
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any

from pyplumio.const import ATTR_EXTRA, ATTR_NAME, ATTR_VALUE, STATE_OFF, STATE_ON
from pyplumio.frames import Request
from pyplumio.helpers.factory import factory
from pyplumio.helpers.schedule import _collect_schedule_data
from pyplumio.helpers.typing import ParameterDataType, ParameterValueType

if TYPE_CHECKING:
    from pyplumio.devices import Device


def _normalize_parameter_value(value: ParameterValueType) -> int:
    """Normalize parameter value to integer."""
    if isinstance(value, str):
        return 1 if value == STATE_ON else 0

    if isinstance(value, tuple):
        # Value is parameter tuple.
        value = value[0]

    return int(value)


def is_binary_parameter(parameter: ParameterDataType) -> bool:
    """Check if parameter is binary."""
    _, min_value, max_value = parameter
    return min_value == 0 and max_value == 1


class Parameter(ABC):
    """Represents device parameter."""

    device: Device
    name: str
    extra: Any
    _value: int
    _min_value: int
    _max_value: int
    _changed: bool = False

    def __init__(
        self,
        device: Device,
        name: str,
        value: ParameterValueType,
        min_value: ParameterValueType,
        max_value: ParameterValueType,
        extra: Any = None,
    ):
        """Initialize Parameter object."""
        self.device = device
        self.name = name
        self.extra = extra
        self._value = _normalize_parameter_value(value)
        self._min_value = _normalize_parameter_value(min_value)
        self._max_value = _normalize_parameter_value(max_value)
        self._changed = False

    def __repr__(self) -> str:
        """Returns serializable string representation."""
        return (
            self.__class__.__name__
            + f"(device={self.device.__class__.__name__}, name={self.name}, "
            + f"value={self._value}, min_value={self._min_value}, "
            + f"max_value={self._max_value}, extra={self.extra})"
        )

    def _call_relational_method(self, method_to_call, other):
        func = getattr(self.value, method_to_call)
        return func(_normalize_parameter_value(other))

    def __int__(self) -> int:
        """Return integer representation of parameter value."""
        return self.value

    def __add__(self, other) -> int:
        """Return result of addition."""
        return self._call_relational_method("__add__", other)

    def __sub__(self, other) -> int:
        """Return result of the subtraction."""
        return self._call_relational_method("__sub__", other)

    def __truediv__(self, other):
        """Return result of true division."""
        return self._call_relational_method("__truediv__", other)

    def __floordiv__(self, other):
        """Return result of floored division."""
        return self._call_relational_method("__floordiv__", other)

    def __mul__(self, other):
        """Return result of the multiplication."""
        return self._call_relational_method("__mul__", other)

    def __eq__(self, other) -> bool:
        """Compare if parameter value is equal to other."""
        return self._call_relational_method("__eq__", other)

    def __ge__(self, other) -> bool:
        """Compare if parameter value is greater or equal to other."""
        return self._call_relational_method("__ge__", other)

    def __gt__(self, other) -> bool:
        """Compare if parameter value is greater than other."""
        return self._call_relational_method("__gt__", other)

    def __le__(self, other) -> bool:
        """Compare if parameter value is less or equal to other."""
        return self._call_relational_method("__le__", other)

    def __lt__(self, other) -> bool:
        """Compare if parameter value is less that other."""
        return self._call_relational_method("__lt__", other)

    def set(self, value: ParameterValueType) -> None:
        """Set parameter value."""
        value = _normalize_parameter_value(value)
        if value == self._value:
            return

        if self.min_value <= value <= self.max_value:
            self._value = value
            self._changed = True
            self.device.queue.put_nowait(self.request)
            return

        raise ValueError(
            f"Parameter value must be between '{self.min_value}' and '{self.max_value}'"
        )

    @property
    def value(self) -> int:
        """Return parameter value."""
        return self._value

    @property
    def min_value(self) -> int:
        """Return minimum allowed value."""
        return self._min_value

    @property
    def max_value(self) -> int:
        """Return maximum allowed value."""
        return self._max_value

    @property
    def changed(self) -> bool:
        """Has parameter been changed recently."""
        return self._changed

    @property
    @abstractmethod
    def request(self) -> Request:
        """Return request to change the parameter."""


class BinaryParameter(Parameter):
    """Represents binary device parameter."""

    def turn_on(self) -> None:
        """Turn parameter on."""
        self.set(STATE_ON)

    def turn_off(self) -> None:
        """Turn parameter off"""
        self.set(STATE_OFF)


class BoilerParameter(Parameter):
    """Represents boiler parameter."""

    @property
    def request(self) -> Request:
        """Return request to change the parameter."""
        handler = (
            "frames.requests.BoilerControlRequest"
            if self.name == "boiler_control"
            else "frames.requests.SetBoilerParameterRequest"
        )

        return factory(
            handler,
            recipient=self.device.address,
            data={ATTR_NAME: self.name, ATTR_VALUE: self.value},
        )


class BoilerBinaryParameter(BoilerParameter, BinaryParameter):
    """Represents boiler binary parameter."""


class MixerParameter(Parameter):
    """Represents mixer parameter."""

    @property
    def request(self) -> Request:
        """Return request to change the parameter."""
        return factory(
            "frames.requests.SetMixerParameterRequest",
            recipient=self.device.address,
            data={ATTR_NAME: self.name, ATTR_VALUE: self.value, ATTR_EXTRA: self.extra},
        )


class MixerBinaryParameter(MixerParameter, BinaryParameter):
    """Represents mixer binary parameter."""


class ScheduleParameter(Parameter):
    """Represents schedule parameter."""

    @property
    def request(self) -> Request:
        """Return request to change the parameter."""
        return factory(
            "frames.requests.SetScheduleRequest",
            recipient=self.device.address,
            data=_collect_schedule_data(self.extra, self.device),
        )


class ScheduleBinaryParameter(ScheduleParameter, BinaryParameter):
    """Represents schedule binary parameter."""



"""Contains regulator parameter structure decoder."""
from __future__ import annotations

from typing import Dict, Final, List, Optional, Tuple

from pyplumio import util
from pyplumio.const import ATTR_BOILER_PARAMETERS
from pyplumio.helpers.typing import DeviceDataType, ParameterDataType
from pyplumio.structures import StructureDecoder, make_device_data

PARAMETER_BOILER_CONTROL: Final = "boiler_control"
BOILER_PARAMETERS: List[str] = [
    "airflow_power_100",
    "airflow_power_50",
    "airflow_power_30",
    "power_100",
    "power_50",
    "power_30",
    "max_fan_boiler_power",
    "min_fan_boiler_power",
    "fuel_feeding_time_100",
    "fuel_feeding_time_50",
    "fuel_feeding_time_30",
    "fuel_feeding_break_100",
    "fuel_feeding_break_50",
    "fuel_feeding_break_30",
    "cycle_time",
    "h2_hysteresis",
    "h1_hysteresis",
    "heating_hysteresis",
    "fuzzy_logic",
    "min_fuzzy_logic_power",
    "max_fuzzy_logic_power",
    "min_boiler_power",
    "max_boiler_power",
    "min_fan_power",
    "max_fan_power",
    "t_reduction_airflow",
    "fan_power_gain",
    "fuel_flow_correction_fuzzy_logic",
    "fuel_flow_correction",
    "airflow_correction_100",
    "feeder_correction_100",
    "airflow_correction_50",
    "feeder_correction_50",
    "airflow_correction_30",
    "feeder_correction_30",
    "airflow_power_grate",
    "boiler_hysteresis_grate",
    "supervision_work_airflow",
    "supervision_work_airflow_brake",
    "heating_temp_grate",
    "fuel_detection_time_grate",
    "airflow_power_kindle",
    "small_airflow_power_kindle",
    "airflow_kindle_delay",
    "scavenge_kindle",
    "feeder_kindle",
    "feeder_kindle_weight",
    "kindle_time",
    "warming_up_time",
    "kindle_finish_fumes_temp",
    "kindle_finish_threshold",
    "kindle_fumes_delta_temp",
    "kindle_delta_t",
    "min_kindle_power_time",
    "scavenge_after_kindle",
    "airflow_power_after_kindle",
    "supervision_time",
    "feed_time_supervision",
    "feed_time_supervision_weight",
    "feed_supervision_break",
    "supervision_cycle_duration",
    "airflow_power_supervision",
    "fan_supervison_break",
    "fan_work_supervision",
    "increase_fan_support_mode",
    "max_extinguish_time",
    "min_extinguish_time",
    "extinguish_time",
    "airflow_power_extinguish",
    "airflow_work_extinguish",
    "airflow_brake_extinguish",
    "scavenge_start_extinguish",
    "scavenge_stop_extinguish",
    "clean_begin_time",
    "extinguish_clean_time",
    "airflow_power_clean",
    "warming_up_brake_time",
    "warming_up_cycle_time",
    "remind_time",
    "lambda_work",
    "lambda_correction_range",
    "oxygen_100",
    "oxygen_50",
    "oxygen_30",
    "oxygen_correction_fl",
    "fuel_flow_kg_h",
    "feeder_calibration",
    "fuel_factor",
    "fuel_calorific_value_kwh_kg",
    "fuel_detection_time",
    "fuel_detection_fumes_temp",
    "schedule_feeder_2",
    "feed2_h1",
    "feed2_h2",
    "feed2_h3",
    "feed2_h4",
    "feed2_work",
    "feed2_break",
    "heating_target_temp",
    "min_heating_target_temp",
    "max_heating_target_temp",
    "heating_pump_on_temp",
    "pause_heating_for_water_heater",
    "pause_term",
    "work_term",
    "increase_heating_temp_for_water_heater",
    "heating_weather_control",
    "heating_heat_curve",
    "heating_heat_curve_shift",
    "weather_factor",
    "term_boiler_operation",
    "term_boiler_mode",
    "decrease_set_heating_term",
    "term_pump_off",
    "al_boiler_temp",
    "max_feeder_temp",
    "external_boiler_temp",
    "alarm_notify",
    "pump_hysteresis",
    "water_heater_target_temp",
    "min_water_heater_target_temp",
    "max_water_heater_target_temp",
    "water_heater_work_mode",
    "water_heater_hysteresis",
    "water_heater_disinfection",
    "summer_mode",
    "summer_mode_on_temp",
    "summer_mode_off_temp",
    "water_heater_feeding_extension",
    "circulation_control",
    "circulation_pause_time",
    "circulation_work_time",
    "circulation_start_temp",
    "buffer_control",
    "max_buffer_temp",
    "min_buffer_temp",
    "buffer_hysteresis",
    "buffer_load_start",
    "buffer_load_stop",
]


class BoilerParametersStructure(StructureDecoder):
    """Represents boiler parameters data structure."""

    def decode(
        self, message: bytearray, offset: int = 0, data: Optional[DeviceDataType] = None
    ) -> Tuple[DeviceDataType, int]:
        """Decode bytes and return message data and offset."""
        first_parameter = message[offset + 1]
        parameters_number = message[offset + 2]
        offset += 3
        boiler_parameters: Dict[str, ParameterDataType] = {}
        for index in range(first_parameter, parameters_number + first_parameter):
            parameter = util.unpack_parameter(message, offset)
            if parameter is not None:
                parameter_name = BOILER_PARAMETERS[index]
                boiler_parameters[parameter_name] = parameter

            offset += 3

        return (
            make_device_data(data, {ATTR_BOILER_PARAMETERS: boiler_parameters}),
            offset,
        )



"""Contains data schema structure decoder."""
from __future__ import annotations

from typing import Dict, List, Optional, Tuple

from pyplumio import util
from pyplumio.const import ATTR_MODE, ATTR_SCHEMA
from pyplumio.helpers.data_types import DATA_TYPES, DataType
from pyplumio.helpers.typing import DeviceDataType
from pyplumio.structures import StructureDecoder, make_device_data
from pyplumio.structures.outputs import (
    FAN_OUTPUT,
    FEEDER_OUTPUT,
    HEATING_PUMP_OUTPUT,
    LIGHTER_OUTPUT,
    WATER_HEATER_PUMP_OUTPUT,
)
from pyplumio.structures.statuses import HEATING_TARGET, WATER_HEATER_TARGET
from pyplumio.structures.temperatures import (
    EXHAUST_TEMP,
    FEEDER_TEMP,
    HEATING_TEMP,
    OUTSIDE_TEMP,
    WATER_HEATER_TEMP,
)

REGDATA_SCHEMA: Dict[int, str] = {
    1792: ATTR_MODE,
    1024: HEATING_TEMP,
    1026: FEEDER_TEMP,
    1025: WATER_HEATER_TEMP,
    1027: OUTSIDE_TEMP,
    1030: EXHAUST_TEMP,
    1280: HEATING_TARGET,
    1281: WATER_HEATER_TARGET,
    1536: FAN_OUTPUT,
    1538: FEEDER_OUTPUT,
    1541: HEATING_PUMP_OUTPUT,
    1542: WATER_HEATER_PUMP_OUTPUT,
    3: LIGHTER_OUTPUT,
}


class DataSchemaStructure(StructureDecoder):
    """Represent data schema structure."""

    def decode(
        self, message: bytearray, offset: int = 0, data: Optional[DeviceDataType] = None
    ) -> Tuple[DeviceDataType, int]:
        """Decode bytes and return message data and offset."""
        blocks_number = util.unpack_ushort(message[offset : offset + 2])
        offset += 2
        schema: List[Tuple[str, DataType]] = []
        if blocks_number > 0:
            for _ in range(blocks_number):
                param_type = message[offset]
                param_id = util.unpack_ushort(message[offset + 1 : offset + 3])
                param_name = REGDATA_SCHEMA.get(param_id, str(param_id))
                schema.append((param_name, DATA_TYPES[param_type]()))
                offset += 3

        return make_device_data(data, {ATTR_SCHEMA: schema}), offset



"""Contains network program version decoder."""

import struct
from typing import Optional, Tuple

from pyplumio.const import ATTR_VERSION
from pyplumio.helpers.typing import DeviceDataType
from pyplumio.helpers.version_info import VersionInfo
from pyplumio.structures import Structure, make_device_data


class ProgramVersionStructure(Structure):
    """Represents program version data structure."""

    def encode(self, data: DeviceDataType) -> bytearray:
        """Encode device data to bytearray message."""
        message = bytearray(15)
        version_info = data[ATTR_VERSION] if ATTR_VERSION in data else VersionInfo()
        struct.pack_into(
            "<2sB2s3s3HB",
            message,
            0,
            version_info.struct_tag,
            version_info.struct_version,
            version_info.device_id,
            version_info.processor_signature,
            *map(int, version_info.software.split(".", 2)),
            self.frame.sender,
        )
        return message

    def decode(
        self, message: bytearray, offset: int = 0, data: Optional[DeviceDataType] = None
    ) -> Tuple[DeviceDataType, int]:
        """Decode bytes and return message data and offset."""
        version_info = VersionInfo()
        [
            version_info.struct_tag,
            version_info.struct_version,
            version_info.device_id,
            version_info.processor_signature,
            software_version1,
            software_version2,
            software_version3,
            self.frame.recipient,
        ] = struct.unpack_from("<2sB2s3s3HB", message)
        version_info.software = ".".join(
            map(str, [software_version1, software_version2, software_version3])
        )

        return make_device_data(data, {ATTR_VERSION: version_info}), offset + 15



"""Contains outputs structure decoder."""
from __future__ import annotations

import math
from typing import Final, Optional, Tuple

from pyplumio import util
from pyplumio.helpers.typing import DeviceDataType
from pyplumio.structures import StructureDecoder, make_device_data

FAN_OUTPUT: Final = "fan"
FEEDER_OUTPUT: Final = "feeder"
HEATING_PUMP_OUTPUT: Final = "heating_pump"
WATER_HEATER_PUMP_OUTPUT: Final = "water_heater_pump"
CIRCULATION_PUMP_OUTPUT: Final = "ciculation_pump"
LIGHTER_OUTPUT: Final = "lighter"
ALARM_OUTPUT: Final = "alarm"
OUTER_BOILER_OUTPUT: Final = "outer_boiler"
FAN2_EXHAUST_OUTPUT: Final = "fan2_exhaust"
FEEDER2_OUTPUT: Final = "feeder2"
OUTER_FEEDER_OUTPUT: Final = "outer_feeder"
SOLAR_PUMP_OUTPUT: Final = "solar_pump"
FIREPLACE_PUMP_OUTPUT: Final = "fireplace_pump"
GCZ_CONTACT: Final = "gcz_contact"
BLOW_FAN1_OUTPUT: Final = "blow_fan1"
BLOW_FAN2_OUTPUT: Final = "blow_fan2"
OUTPUTS: Tuple[str, ...] = (
    FAN_OUTPUT,
    FEEDER_OUTPUT,
    HEATING_PUMP_OUTPUT,
    WATER_HEATER_PUMP_OUTPUT,
    CIRCULATION_PUMP_OUTPUT,
    LIGHTER_OUTPUT,
    ALARM_OUTPUT,
    OUTER_BOILER_OUTPUT,
    FAN2_EXHAUST_OUTPUT,
    FEEDER2_OUTPUT,
    OUTER_FEEDER_OUTPUT,
    SOLAR_PUMP_OUTPUT,
    FIREPLACE_PUMP_OUTPUT,
    GCZ_CONTACT,
    BLOW_FAN1_OUTPUT,
    BLOW_FAN2_OUTPUT,
)


class OutputsStructure(StructureDecoder):
    """Represent outputs data structure."""

    def decode(
        self, message: bytearray, offset: int = 0, data: Optional[DeviceDataType] = None
    ) -> Tuple[DeviceDataType, int]:
        """Decode bytes and return message data and offset."""
        outputs = util.unpack_ushort(message[offset : offset + 4])
        data = make_device_data(data)
        for index, output in enumerate(OUTPUTS):
            data[output] = bool(outputs & int(math.pow(2, index)))

        return data, offset + 4



"""Contains protocol representation."""
from __future__ import annotations

import asyncio
import logging
from typing import Awaitable, Callable, Dict, Final, Optional, Tuple

from pyplumio.const import ATTR_NETWORK
from pyplumio.devices import Device, DeviceTypes, get_device_handler
from pyplumio.exceptions import (
    FrameError,
    ReadError,
    UnknownDeviceError,
    UnknownFrameError,
)
from pyplumio.frames import FrameTypes
from pyplumio.frames.requests import StartMasterRequest
from pyplumio.helpers.factory import factory
from pyplumio.helpers.network_info import (
    EthernetParameters,
    NetworkInfo,
    WirelessParameters,
)
from pyplumio.helpers.task_manager import TaskManager
from pyplumio.stream import FrameReader, FrameWriter

_LOGGER = logging.getLogger(__name__)

CONSUMERS_NUMBER: Final = 2


def _get_device_handler_and_name(address: int) -> Tuple[str, str]:
    """Get device handler full path and lowercased class name."""
    handler = get_device_handler(address)
    _, class_name = handler.rsplit(".", 1)
    return handler, class_name.lower()


class Protocol(TaskManager):
    """Represents protocol."""

    writer: Optional[FrameWriter]
    reader: Optional[FrameReader]
    devices: Dict[str, Device]
    connected: asyncio.Event
    _network: NetworkInfo
    _queues: Tuple[asyncio.Queue, asyncio.Queue]
    _connection_lost_callback: Optional[Callable[[], Awaitable[None]]]

    def __init__(
        self,
        connection_lost_callback: Optional[Callable[[], Awaitable[None]]] = None,
        ethernet_parameters: Optional[EthernetParameters] = None,
        wireless_parameters: Optional[WirelessParameters] = None,
    ):
        """Initialize new Protocol object."""
        super().__init__()
        self.writer = None
        self.reader = None
        self.devices = {}
        self.connected = asyncio.Event()
        read_queue: asyncio.Queue = asyncio.Queue()
        write_queue: asyncio.Queue = asyncio.Queue()
        self._queues = (read_queue, write_queue)
        self._connection_lost_callback = connection_lost_callback
        if ethernet_parameters is None:
            ethernet_parameters = EthernetParameters()

        if wireless_parameters is None:
            wireless_parameters = WirelessParameters()

        self._network = NetworkInfo(eth=ethernet_parameters, wlan=wireless_parameters)

    def __getattr__(self, name: str):
        """Return attributes from the underlying devices table."""
        if name in self.devices:
            return self.devices[name]

        raise AttributeError

    async def frame_producer(
        self, read_queue: asyncio.Queue, lock: asyncio.Lock
    ) -> None:
        """Handle frame reads."""
        await self.connected.wait()
        while self.connected.is_set():
            await lock.acquire()
            try:
                frame = await self.reader.read()
                if frame is not None:
                    read_queue.put_nowait(frame)
            except UnknownFrameError as e:
                _LOGGER.debug("Unknown frame type: %s", e)
            except ReadError as e:
                _LOGGER.debug("Read error: %s", e)
            except FrameError as e:
                _LOGGER.warning("Can't process received frame: %s", e)
            except (OSError, asyncio.TimeoutError):
                self.create_task(self.connection_lost())
                break
            finally:
                lock.release()

    async def write_consumer(
        self, write_queue: asyncio.Queue, lock: asyncio.Lock
    ) -> None:
        """Handle frame writes."""
        await self.connected.wait()
        while self.connected.is_set():
            frame = await write_queue.get()
            await lock.acquire()
            try:
                await self.writer.write(frame)
            except (OSError, asyncio.TimeoutError):
                self.create_task(self.connection_lost())
                break
            finally:
                write_queue.task_done()
                lock.release()

    async def frame_consumer(
        self, read_queue: asyncio.Queue, write_queue: asyncio.Queue
    ) -> None:
        """Handle frame processing."""
        await self.connected.wait()
        while self.connected.is_set():
            frame = await read_queue.get()
            try:
                handler, name = _get_device_handler_and_name(frame.sender)
            except UnknownDeviceError as e:
                _LOGGER.debug("Unknown device: %s", e)
                read_queue.task_done()
                continue

            if name not in self.devices:
                device = factory(handler, queue=write_queue)
                self.devices[name] = device
                self.set_event(name)

            self.devices[name].handle_frame(frame)

            if frame.frame_type in (
                FrameTypes.REQUEST_CHECK_DEVICE,
                FrameTypes.REQUEST_PROGRAM_VERSION,
            ):
                write_queue.put_nowait(
                    frame.response(data={ATTR_NETWORK: self._network})
                )

            read_queue.task_done()

    def connection_established(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
    ):
        """Start consumers and producer tasks."""
        self.reader = FrameReader(reader)
        self.writer = FrameWriter(writer)
        read_queue, write_queue = self.queues
        write_queue.put_nowait(StartMasterRequest(recipient=DeviceTypes.ECOMAX))
        lock: asyncio.Lock = asyncio.Lock()
        self.create_task(self.write_consumer(write_queue, lock))
        self.create_task(self.frame_producer(read_queue, lock))
        for _ in range(CONSUMERS_NUMBER):
            self.create_task(self.frame_consumer(*self.queues))

        self.connected.set()

    async def connection_lost(self):
        """Shutdown consumers and call connection lost callback."""
        self.connected.clear()
        if self._connection_lost_callback is not None:
            await self._connection_lost_callback()

    async def shutdown(self):
        """Shutdown protocol tasks."""
        await asyncio.wait([queue.join() for queue in self.queues])
        self.cancel_tasks()
        await self.wait_until_done()

        for device in self.devices.values():
            await device.shutdown()

        if self.writer:
            await self.writer.close()

    async def get_device(self, device: str, timeout: Optional[float] = None) -> Device:
        """Wait for device and return it once it's available."""
        if device not in self.devices:
            await asyncio.wait_for(self.create_event(device).wait(), timeout=timeout)

        return self.devices[device]

    @property
    def queues(self) -> Tuple[asyncio.Queue, asyncio.Queue]:
        """Return protocol queues."""
        return self._queues



"""Contains network information dataclasses."""
from __future__ import annotations

from dataclasses import dataclass
from typing import Final

WLAN_ENCRYPTION_UNKNOWN: Final = 0
WLAN_ENCRYPTION_NONE: Final = 1
WLAN_ENCRYPTION_WEP: Final = 2
WLAN_ENCRYPTION_WPA: Final = 3
WLAN_ENCRYPTION_WPA2: Final = 4
WLAN_ENCRYPTION: Final = (
    WLAN_ENCRYPTION_UNKNOWN,
    WLAN_ENCRYPTION_NONE,
    WLAN_ENCRYPTION_WEP,
    WLAN_ENCRYPTION_WPA,
    WLAN_ENCRYPTION_WPA2,
)

DEFAULT_IP: Final = "0.0.0.0"
DEFAULT_NETMASK: Final = "255.255.255.0"


@dataclass
class EthernetParameters:
    """Represents ethernet parameters."""

    ip: str = DEFAULT_IP
    netmask: str = DEFAULT_NETMASK
    gateway: str = DEFAULT_IP
    status: bool = False


@dataclass
class WirelessParameters(EthernetParameters):
    """Represents wireless network parameters."""

    ssid: str = ""
    encryption: int = WLAN_ENCRYPTION_NONE
    signal_quality: int = 100


@dataclass
class NetworkInfo:
    """Represents network parameters."""

    eth: EthernetParameters = EthernetParameters()
    wlan: WirelessParameters = WirelessParameters()
    server_status: bool = True



"""Contains lambda structure decoder."""
from __future__ import annotations

import math
from typing import Final, Optional, Tuple

from pyplumio import util
from pyplumio.const import ATTR_LAMBDA_SENSOR
from pyplumio.helpers.typing import DeviceDataType
from pyplumio.structures import StructureDecoder, make_device_data

LAMBDA_LEVEL: Final = "lambda_level"
LAMBDA_STATUS: Final = "lambda_status"
LAMBDA_TARGET: Final = "lambda_target"
LAMBDA: Tuple[str, ...] = (
    LAMBDA_LEVEL,
    LAMBDA_STATUS,
    LAMBDA_TARGET,
)


class LambaSensorStructure(StructureDecoder):
    """Represents lambda sensor data structure."""

    def decode(
        self, message: bytearray, offset: int = 0, data: Optional[DeviceDataType] = None
    ) -> Tuple[DeviceDataType, int]:
        """Decode bytes and return message data and offset."""
        lambda_sensor: DeviceDataType = {}
        if message[offset] != 0xFF:
            lambda_sensor[LAMBDA_STATUS] = message[offset]
            lambda_sensor[LAMBDA_TARGET] = message[offset + 1]
            lambda_level = util.unpack_ushort(message[offset + 2 : offset + 4])
            lambda_sensor[LAMBDA_LEVEL] = (
                None if math.isnan(lambda_level) else lambda_level
            )
            offset += 3

        return make_device_data(data, {ATTR_LAMBDA_SENSOR: lambda_sensor}), offset + 1



"""Contains type aliases."""
from __future__ import annotations

from typing import Any, Awaitable, Callable, Dict, Literal, Tuple, Union

NumericType = Union[int, float]
ParameterDataType = Tuple[int, int, int]
ParameterValueType = Union[NumericType, bool, Literal["off"], Literal["on"]]
DeviceDataType = Dict[str, Any]
SensorCallbackType = Callable[[Any], Awaitable[Any]]
VersionsInfoType = Dict[int, int]
BytesType = Union[bytes, bytearray]
MessageType = bytearray



"""Contains request frames."""
from __future__ import annotations

from typing import ClassVar, Optional

from pyplumio.const import ATTR_EXTRA, ATTR_NAME, ATTR_VALUE
from pyplumio.exceptions import FrameDataError
from pyplumio.frames import FrameTypes, Request, Response
from pyplumio.frames.responses import DeviceAvailableResponse, ProgramVersionResponse
from pyplumio.helpers.typing import DeviceDataType, MessageType
from pyplumio.structures.boiler_parameters import BOILER_PARAMETERS
from pyplumio.structures.mixer_parameters import MIXER_PARAMETERS
from pyplumio.structures.schedules import SchedulesStructure


class ProgramVersionRequest(Request):
    """Represents program version request."""

    frame_type: ClassVar[int] = FrameTypes.REQUEST_PROGRAM_VERSION

    def response(self, **kwargs) -> Optional[Response]:
        """Return response frame object."""
        return ProgramVersionResponse(recipient=self.sender, **kwargs)


class CheckDeviceRequest(Request):
    """Represents check device request."""

    frame_type: ClassVar[int] = FrameTypes.REQUEST_CHECK_DEVICE

    def response(self, **kwargs) -> Optional[Response]:
        """Return response frame object."""
        return DeviceAvailableResponse(recipient=self.sender, **kwargs)


class UIDRequest(Request):
    """Represents uid request."""

    frame_type: ClassVar[int] = FrameTypes.REQUEST_UID


class PasswordRequest(Request):
    """Represents password request."""

    frame_type: ClassVar[int] = FrameTypes.REQUEST_PASSWORD


class BoilerParametersRequest(Request):
    """Represents boiler parameters request."""

    frame_type: ClassVar[int] = FrameTypes.REQUEST_BOILER_PARAMETERS

    def create_message(self, data: DeviceDataType) -> MessageType:
        """Create frame message."""
        message = bytearray()
        message.append(255)  # Number of parameters.
        message.append(0)  # Index of the first parameter.
        return message


class MixerParametersRequest(Request):
    """Represents mixer parameters request."""

    frame_type: ClassVar[int] = FrameTypes.REQUEST_MIXER_PARAMETERS


class DataSchemaRequest(Request):
    """Represents data schema request."""

    frame_type: ClassVar[int] = FrameTypes.REQUEST_DATA_SCHEMA


class SetBoilerParameterRequest(Request):
    """Represents set boiler parameter request."""

    frame_type: ClassVar[int] = FrameTypes.REQUEST_SET_BOILER_PARAMETER

    def create_message(self, data: DeviceDataType) -> MessageType:
        """Create frame message."""
        try:
            message = bytearray()
            name = data[ATTR_NAME]
            value = data[ATTR_VALUE]
            message.append(BOILER_PARAMETERS.index(name))
            message.append(value)
            return message
        except (KeyError, ValueError) as e:
            raise FrameDataError from e


class SetMixerParameterRequest(Request):
    """Represents set mixer parameter request."""

    frame_type: ClassVar[int] = FrameTypes.REQUEST_SET_MIXER_PARAMETER

    def create_message(self, data: DeviceDataType) -> MessageType:
        """Create frame message."""
        try:
            name = data[ATTR_NAME]
            value = data[ATTR_VALUE]
            index = data[ATTR_EXTRA]
            message = bytearray()
            message.append(index)
            message.append(MIXER_PARAMETERS.index(name))
            message.append(value)
            return message
        except (KeyError, ValueError) as e:
            raise FrameDataError from e


class BoilerControlRequest(Request):
    """Represent boiler control request."""

    frame_type: ClassVar[int] = FrameTypes.REQUEST_BOILER_CONTROL

    def create_message(self, data: DeviceDataType) -> MessageType:
        """Creates frame message."""
        try:
            message = bytearray()
            message.append(data[ATTR_VALUE])
            return message
        except KeyError as e:
            raise FrameDataError from e


class StartMasterRequest(Request):
    """Represent start master request."""

    frame_type: ClassVar[int] = FrameTypes.REQUEST_START_MASTER


class StopMasterRequest(Request):
    """Represent stop master request."""

    frame_type: ClassVar[int] = FrameTypes.REQUEST_STOP_MASTER


class AlertsRequest(Request):
    """Represent alerts request."""

    frame_type: ClassVar[int] = FrameTypes.REQUEST_ALERTS

    def create_message(self, data: DeviceDataType) -> MessageType:
        """Create frame message."""
        message = bytearray()
        message.append(0)  # Index of the first alarm.
        message.append(100)  # Number of alarms.
        return message


class SchedulesRequest(Request):
    """Represents schedule request."""

    frame_type: ClassVar[int] = FrameTypes.REQUEST_SCHEDULES


class SetScheduleRequest(Request):
    """Represents set schedule request."""

    frame_type: ClassVar[int] = FrameTypes.REQUEST_SET_SCHEDULE

    def create_message(self, data: DeviceDataType) -> MessageType:
        return SchedulesStructure(self).encode(data)



"""Contains timeout decorator."""
from __future__ import annotations

import asyncio
import logging

_LOGGER = logging.getLogger(__name__)


def timeout(seconds: int, raise_exception: bool = True):
    """Decorator to add a timeout to the awaitable. Returns None if
    raise_exception parameter is set to false."""

    def decorator(func):
        async def wrapper(*args, **kwargs):
            try:
                return await asyncio.wait_for(func(*args, **kwargs), timeout=seconds)
            except asyncio.TimeoutError:
                if raise_exception:
                    raise

                _LOGGER.warning(
                    "Function '%s' timed out after %i seconds",
                    func.__name__,
                    seconds,
                )
                return None

        return wrapper

    return decorator



"""Contains outputs structure decoder."""
from __future__ import annotations

from typing import Final, Optional, Tuple

from pyplumio.helpers.typing import DeviceDataType
from pyplumio.structures import StructureDecoder, make_device_data

HEATING_TARGET: Final = "heating_target"
HEATING_STATUS: Final = "heating_status"
WATER_HEATER_TARGET: Final = "water_heater_target"
WATER_HEATER_STATUS: Final = "water_heater_status"
STATUSES: Tuple[str, ...] = (
    HEATING_TARGET,
    HEATING_STATUS,
    WATER_HEATER_TARGET,
    WATER_HEATER_STATUS,
)


class StatusesStructure(StructureDecoder):
    """Represents statuses data structure."""

    def decode(
        self, message: bytearray, offset: int = 0, data: Optional[DeviceDataType] = None
    ) -> Tuple[DeviceDataType, int]:
        """Decode bytes and return message data and offset."""
        data = make_device_data(data)
        for index, status in enumerate(STATUSES):
            data[status] = message[offset + index]

        return data, offset + 4



"""Contains callback filters."""
from __future__ import annotations

from abc import ABC, abstractmethod
import math
import time
from typing import Any, Final, Optional

from pyplumio.helpers.parameter import Parameter
from pyplumio.helpers.typing import NumericType, SensorCallbackType

TOLERANCE: Final = 0.1


def _significantly_changed(old_value, new_value) -> bool:
    """Check if value is significantly changed."""
    if old_value is None or (isinstance(old_value, Parameter) and old_value.changed):
        return True

    if isinstance(old_value, (int, float)) and isinstance(new_value, (int, float)):
        return not math.isclose(old_value, new_value, abs_tol=TOLERANCE)

    return old_value != new_value


def _diffence_between(old_value, new_value):
    """Return the difference between values."""
    if isinstance(old_value, list) and isinstance(new_value, list):
        return [x for x in new_value if x not in old_value]

    if hasattr(old_value, "__sub__") and hasattr(new_value, "__sub__"):
        return new_value - old_value

    return None


class Filter(ABC):
    """Represents base for value callback modifiers."""

    _callback: Any
    _value: Any

    def __init__(self, callback: SensorCallbackType):
        """Initialize new Filter object."""
        self._callback = callback
        self._value = None

    def __eq__(self, other) -> bool:
        """Compare debounced callbacks."""
        if isinstance(other, Filter):
            return self._callback == other._callback

        if callable(other):
            return self._callback == other

        raise TypeError

    @abstractmethod
    async def __call__(self, new_value):
        """Set new value for the callback."""


class _OnChange(Filter):
    """Provides changed functionality to the callback."""

    async def __call__(self, new_value):
        """Set new value for the callback."""
        if _significantly_changed(self._value, new_value):
            self._value = new_value
            return await self._callback(new_value)


def on_change(callback: SensorCallbackType) -> _OnChange:
    """Helper for change callback filter."""
    return _OnChange(callback)


class _Debounce(Filter):
    """Provides debounce functionality to the callback."""

    _calls: int = 0
    _min_calls: int = 3

    def __init__(self, callback: SensorCallbackType, min_calls: int):
        """Initialize Debounce object."""
        super().__init__(callback)
        self._calls = 0
        self._min_calls = min_calls

    async def __call__(self, new_value):
        """Set new value for the callback."""
        if _significantly_changed(self._value, new_value):
            self._calls += 1
        else:
            self._calls = 0

        if self._calls >= self._min_calls or self._value is None:
            self._value = new_value
            self._calls = 0
            return await self._callback(new_value)


def debounce(callback: SensorCallbackType, min_calls) -> _Debounce:
    """Helper method for debounce callback filter."""
    return _Debounce(callback, min_calls)


class _Throttle(Filter):
    """Provides throttle functionality to the callback."""

    _last_called: Optional[float]
    _timeout: float

    def __init__(self, callback: SensorCallbackType, seconds: float):
        """Initialize Debounce object."""
        super().__init__(callback)
        self._last_called = None
        self._timeout = seconds

    async def __call__(self, new_value):
        """Set new value for the callback."""
        current_timestamp = time.time()
        if (
            self._last_called is None
            or (current_timestamp - self._last_called) >= self._timeout
        ):
            self._last_called = current_timestamp
            return await self._callback(new_value)


def throttle(callback: SensorCallbackType, seconds: float) -> _Throttle:
    """Helper method for throttle callback filter."""
    return _Throttle(callback, seconds)


class _Delta(Filter):
    """Provides ability to pass call difference to the callback."""

    async def __call__(self, new_value):
        """Set new value for the callback."""
        old_value = self._value
        if _significantly_changed(old_value, new_value):
            self._value = new_value
            if (difference := _diffence_between(old_value, new_value)) is not None:
                return await self._callback(difference)


def delta(callback: SensorCallbackType) -> _Delta:
    """Helper method for delta callback filter."""
    return _Delta(callback)


class _Aggregate(Filter):
    """Provides ability to sum value for some time before sending them
    to the callback."""

    _sum: NumericType
    _last_update: Optional[float]
    _timeout: float

    def __init__(self, callback: SensorCallbackType, seconds: float):
        """Initialize Aggregate object."""
        super().__init__(callback)
        self._last_update = time.time()
        self._timeout = seconds
        self._sum = 0.0

    async def __call__(self, new_value):
        """Set new value for the callback."""
        if not isinstance(new_value, (int, float)):
            raise ValueError("Aggregate filter can't be used for non-numeric values")

        current_timestamp = time.time()
        self._sum += new_value
        if current_timestamp - self._last_update >= self._timeout:
            result = await self._callback(self._sum)
            self._last_update = current_timestamp
            self._sum = 0
            return result


def aggregate(callback: SensorCallbackType, seconds: float) -> _Aggregate:
    """Helper method for total callback filter."""
    return _Aggregate(callback, seconds)



"""Contains ecoMAX device representation."""
from __future__ import annotations

import asyncio
from collections.abc import Sequence
import time
from typing import ClassVar, List

from pyplumio.const import (
    ATTR_BOILER_PARAMETERS,
    ATTR_BOILER_SENSORS,
    ATTR_FUEL_BURNED,
    ATTR_FUEL_CONSUMPTION,
    ATTR_MIXER_PARAMETERS,
    ATTR_MIXER_SENSORS,
    ATTR_MIXERS,
    ATTR_MODE,
    ATTR_PARAMETER,
    ATTR_REGDATA,
    ATTR_SCHEDULE,
    ATTR_SCHEDULES,
    ATTR_SCHEMA,
    ATTR_SWITCH,
)
from pyplumio.devices import Device, DeviceTypes, Mixer
from pyplumio.frames import FrameTypes
from pyplumio.helpers.data_types import Boolean
from pyplumio.helpers.filters import on_change
from pyplumio.helpers.parameter import (
    BoilerBinaryParameter,
    BoilerParameter,
    MixerBinaryParameter,
    MixerParameter,
    ScheduleBinaryParameter,
    ScheduleParameter,
    is_binary_parameter,
)
from pyplumio.helpers.schedule import Schedule, ScheduleDay
from pyplumio.helpers.typing import DeviceDataType
from pyplumio.structures.boiler_parameters import PARAMETER_BOILER_CONTROL


class EcoMAX(Device):
    """Represents ecoMAX controller."""

    address: ClassVar[int] = DeviceTypes.ECOMAX
    _fuel_burned_timestamp: float = 0.0
    _required_frames: List[int] = [
        FrameTypes.REQUEST_UID,
        FrameTypes.REQUEST_DATA_SCHEMA,
        FrameTypes.REQUEST_BOILER_PARAMETERS,
        FrameTypes.REQUEST_MIXER_PARAMETERS,
        FrameTypes.REQUEST_PASSWORD,
        FrameTypes.REQUEST_ALERTS,
        FrameTypes.REQUEST_SCHEDULES,
    ]

    def __init__(self, queue: asyncio.Queue):
        """Initialize new ecoMAX object."""
        super().__init__(queue)
        self._fuel_burned_timestamp = time.time()
        self.subscribe(ATTR_BOILER_SENSORS, self._add_boiler_sensors)
        self.subscribe(ATTR_MODE, on_change(self._add_boiler_control_parameter))
        self.subscribe(ATTR_FUEL_CONSUMPTION, self._add_burned_fuel_counter)
        self.subscribe(ATTR_BOILER_PARAMETERS, self._add_boiler_parameters)
        self.subscribe(ATTR_REGDATA, self._decode_regulator_data)
        self.subscribe(ATTR_MIXER_SENSORS, self._set_mixer_sensors)
        self.subscribe(ATTR_MIXER_PARAMETERS, self._set_mixer_parameters)
        self.subscribe(ATTR_SCHEDULES, self._add_schedules_and_schedule_parameters)

    def _get_mixer(self, mixer_number: int, total_mixers: int) -> Mixer:
        """Get or create a new mixer object and add it to the device."""
        mixers = self.data.setdefault(ATTR_MIXERS, [])
        try:
            mixer = mixers[mixer_number]
        except IndexError:
            mixer = Mixer(mixer_number)
            mixers.append(mixer)
            if len(mixers) == total_mixers:
                # All mixers were processed, notify callbacks and getters.
                self.set_device_data(ATTR_MIXERS, mixers)

        return mixer

    async def _add_boiler_sensors(self, sensors: DeviceDataType) -> bool:
        """Add boiler sensors values to the device data."""
        for name, value in sensors.items():
            await self.async_set_device_data(name, value)

        return True

    async def _add_boiler_parameters(self, parameters: DeviceDataType) -> bool:
        """Add Parameter objects to the device data."""
        for name, value in parameters.items():
            cls = (
                BoilerBinaryParameter if is_binary_parameter(value) else BoilerParameter
            )
            parameter = cls(
                device=self,
                name=name,
                value=value[0],
                min_value=value[1],
                max_value=value[2],
            )
            await self.async_set_device_data(name, parameter)

        return True

    async def _set_mixer_sensors(self, sensors: Sequence[DeviceDataType]) -> bool:
        """Set sensor values for the mixer."""
        for mixer_number, mixer_data in enumerate(sensors):
            mixer = self._get_mixer(mixer_number, len(sensors))
            for name, value in mixer_data.items():
                await mixer.async_set_device_data(name, value)

        return True

    async def _set_mixer_parameters(self, parameters: Sequence[DeviceDataType]) -> bool:
        """Set mixer parameters."""
        for mixer_number, mixer_data in enumerate(parameters):
            mixer = self._get_mixer(mixer_number, len(parameters))
            for name, value in mixer_data.items():
                cls = (
                    MixerBinaryParameter
                    if is_binary_parameter(value)
                    else MixerParameter
                )
                parameter = cls(
                    device=self,
                    name=name,
                    value=value[0],
                    min_value=value[1],
                    max_value=value[2],
                    extra=mixer_number,
                )
                await mixer.async_set_device_data(name, parameter)

        return True

    async def _add_boiler_control_parameter(self, mode: int) -> None:
        """Add BoilerControl parameter to the device instance."""
        parameter = BoilerBinaryParameter(
            device=self,
            name=PARAMETER_BOILER_CONTROL,
            value=(mode != 0),
            min_value=0,
            max_value=1,
        )
        await self.async_set_device_data(PARAMETER_BOILER_CONTROL, parameter)

    async def _add_burned_fuel_counter(self, fuel_consumption: int) -> None:
        """Add burned fuel counter."""
        current_timestamp = time.time()
        seconds_passed = current_timestamp - self._fuel_burned_timestamp
        fuel_burned = (fuel_consumption / 3600) * seconds_passed
        self._fuel_burned_timestamp = current_timestamp
        await self.async_set_device_data(ATTR_FUEL_BURNED, fuel_burned)

    async def _decode_regulator_data(self, regulator_data: bytes) -> DeviceDataType:
        """Add sensor values from the regulator data."""
        offset = 0
        boolean_index = 0
        regdata: DeviceDataType = {}
        schema = self.data.get(ATTR_SCHEMA, [])
        for parameter in schema:
            parameter_id, parameter_type = parameter
            if not isinstance(parameter_type, Boolean) and boolean_index > 0:
                offset += 1
                boolean_index = 0

            parameter_type.unpack(regulator_data[offset:])
            if isinstance(parameter_type, Boolean):
                boolean_index = parameter_type.index(boolean_index)

            regdata[parameter_id] = parameter_type.value
            offset += parameter_type.size

        return regdata

    async def _add_schedules_and_schedule_parameters(
        self, schedules: DeviceDataType
    ) -> DeviceDataType:
        """Add schedules and schedule parameters."""
        data: DeviceDataType = {}
        for name, schedule in schedules.items():
            for field in (ATTR_SWITCH, ATTR_PARAMETER):
                key = f"{ATTR_SCHEDULE}_{name}_{field}"
                value = schedule[field]
                cls = (
                    ScheduleBinaryParameter
                    if is_binary_parameter(value)
                    else ScheduleParameter
                )
                parameter = cls(
                    device=self,
                    name=key,
                    value=value[0],
                    min_value=value[1],
                    max_value=value[2],
                    extra=name,
                )
                await self.async_set_device_data(key, parameter)

            data[name] = Schedule(
                name=name,
                device=self,
                monday=ScheduleDay(schedule[ATTR_SCHEDULE][1]),
                tuesday=ScheduleDay(schedule[ATTR_SCHEDULE][2]),
                wednesday=ScheduleDay(schedule[ATTR_SCHEDULE][3]),
                thursday=ScheduleDay(schedule[ATTR_SCHEDULE][4]),
                friday=ScheduleDay(schedule[ATTR_SCHEDULE][5]),
                saturday=ScheduleDay(schedule[ATTR_SCHEDULE][6]),
                sunday=ScheduleDay(schedule[ATTR_SCHEDULE][0]),
            )

        return data

    async def shutdown(self) -> None:
        """Cancel scheduled tasks."""
        mixers = self.data.get(ATTR_MIXERS, [])
        for mixer in mixers:
            await mixer.shutdown()

        await super().shutdown()



"""Contains product info structure decoder."""

import struct
from typing import Optional, Tuple

from pyplumio import util
from pyplumio.const import ATTR_PRODUCT
from pyplumio.helpers.product_info import ProductInfo
from pyplumio.helpers.typing import DeviceDataType
from pyplumio.helpers.uid import unpack_uid
from pyplumio.structures import StructureDecoder, make_device_data


class ProductInfoStructure(StructureDecoder):
    """Represents product info data structure."""

    def decode(
        self, message: bytearray, offset: int = 0, data: Optional[DeviceDataType] = None
    ) -> Tuple[DeviceDataType, int]:
        """Decode bytes and return message data and offset."""
        product_info = ProductInfo()
        product_info.type, product_info.product = struct.unpack_from("<BH", message)
        product_info.uid = unpack_uid(message, offset)
        product_info.logo, product_info.image = struct.unpack_from("<HH", message)
        product_info.model = util.unpack_string(message, offset + 16)

        return make_device_data(data, {ATTR_PRODUCT: product_info}), offset



"""Contains alarms structure decoder."""
from __future__ import annotations

from typing import Optional, Tuple

from pyplumio.const import ATTR_PENDING_ALERTS
from pyplumio.helpers.typing import DeviceDataType
from pyplumio.structures import StructureDecoder, make_device_data


class PendingAlertsStructure(StructureDecoder):
    """Represents current alert data structure."""

    def decode(
        self, message: bytearray, offset: int = 0, data: Optional[DeviceDataType] = None
    ) -> Tuple[DeviceDataType, int]:
        """Decode bytes and return message data and offset."""
        alerts_number = message[offset]
        pending_alerts = [message[offset + i] for i in range(alerts_number)]
        return make_device_data(data, {ATTR_PENDING_ALERTS: pending_alerts}), (
            offset + alerts_number + 1
        )



"""Contains response frames."""
from __future__ import annotations

from typing import ClassVar

from pyplumio.const import ATTR_PASSWORD
from pyplumio.frames import FrameTypes, Response
from pyplumio.helpers.typing import DeviceDataType, MessageType
from pyplumio.structures.alerts import AlertsStructure
from pyplumio.structures.boiler_parameters import BoilerParametersStructure
from pyplumio.structures.data_schema import DataSchemaStructure
from pyplumio.structures.mixer_parameters import MixerParametersStructure
from pyplumio.structures.network_info import NetworkInfoStructure
from pyplumio.structures.product_info import ProductInfoStructure
from pyplumio.structures.program_version import ProgramVersionStructure
from pyplumio.structures.schedules import SchedulesStructure


class ProgramVersionResponse(Response):
    """Represents program version response. Contains software
    version info.
    """

    frame_type: ClassVar[int] = FrameTypes.RESPONSE_PROGRAM_VERSION

    def create_message(self, data: DeviceDataType) -> MessageType:
        """Create frame message."""
        return ProgramVersionStructure(self).encode(data)

    def decode_message(self, message: MessageType) -> DeviceDataType:
        """Decode frame message."""
        return ProgramVersionStructure(self).decode(message)[0]


class DeviceAvailableResponse(Response):
    """Represents device available response. Contains network
    information and status.
    """

    frame_type: ClassVar[int] = FrameTypes.RESPONSE_DEVICE_AVAILABLE

    def create_message(self, data: DeviceDataType) -> MessageType:
        """Create frame message."""
        return NetworkInfoStructure(self).encode(data)

    def decode_message(self, message: MessageType) -> DeviceDataType:
        """Decode frame message."""
        return NetworkInfoStructure(self).decode(message, offset=1)[0]


class UIDResponse(Response):
    """Represents UID response. Contains product and model info."""

    frame_type: ClassVar[int] = FrameTypes.RESPONSE_UID

    def create_message(self, data: DeviceDataType) -> MessageType:
        """Create frame message."""
        return ProductInfoStructure(self).encode(data)

    def decode_message(self, message: MessageType) -> DeviceDataType:
        """Decode frame message."""
        return ProductInfoStructure(self).decode(message, offset=3)[0]


class PasswordResponse(Response):
    """Represent password response. Contains device service password."""

    frame_type: ClassVar[int] = FrameTypes.RESPONSE_PASSWORD

    def decode_message(self, message: MessageType) -> DeviceDataType:
        """Decode frame message."""
        password = message[1:].decode() if message[1:] else None
        return {ATTR_PASSWORD: password}


class BoilerParametersResponse(Response):
    """Represents boiler parameters response. Contains editable boiler
    parameters.
    """

    frame_type: ClassVar[int] = FrameTypes.RESPONSE_BOILER_PARAMETERS

    def create_message(self, data: DeviceDataType) -> MessageType:
        """Create frame message."""
        return BoilerParametersStructure(self).encode(data)

    def decode_message(self, message: MessageType) -> DeviceDataType:
        """Decode frame message."""
        return BoilerParametersStructure(self).decode(message)[0]


class MixerParametersResponse(Response):
    """Represents mixer parameters response. Contains editable mixer
    parameters.
    """

    frame_type: ClassVar[int] = FrameTypes.RESPONSE_MIXER_PARAMETERS

    def decode_message(self, message: MessageType) -> DeviceDataType:
        """Decode frame message."""
        return MixerParametersStructure(self).decode(message)[0]


class DataSchemaResponse(Response):
    """Represents data schema response. Contains schema that describes
    regdata message structure.
    """

    frame_type: ClassVar[int] = FrameTypes.RESPONSE_DATA_SCHEMA

    def decode_message(self, message: MessageType) -> DeviceDataType:
        """Decode frame message."""
        return DataSchemaStructure(self).decode(message)[0]


class SetBoilerParameterResponse(Response):
    """Represents set boiler parameter response. Empty response
    that aknowledges, that boiler parameter was successfully changed.
    """

    frame_type: ClassVar[int] = FrameTypes.RESPONSE_SET_BOILER_PARAMETER


class SetMixerParameterResponse(Response):
    """Represents set mixer parameter response. Empty response
    that aknowledges, that mixer parameter was successfully changed.
    """

    frame_type: ClassVar[int] = FrameTypes.RESPONSE_SET_MIXER_PARAMETER


class BoilerControlResponse(Response):
    """Represents boiler control response. Empty response
    that aknowledges, that boiler control request was successfully
    processed.
    """

    frame_type: ClassVar[int] = FrameTypes.RESPONSE_BOILER_CONTROL


class AlertsResponse(Response):
    """Represents device alerts."""

    frame_type: ClassVar[int] = FrameTypes.RESPONSE_ALERTS

    def decode_message(self, message: MessageType) -> DeviceDataType:
        """Decode frame message."""
        return AlertsStructure(self).decode(message)[0]


class SchedulesResponse(Response):
    """Represents device schedule."""

    frame_type: ClassVar[int] = FrameTypes.RESPONSE_SCHEDULES

    def decode_message(self, message: MessageType) -> DeviceDataType:
        """Decode frame message."""
        return SchedulesStructure(self).decode(message)[0]



"""Contains UID decoder."""
from __future__ import annotations


def unpack_uid(message: bytearray, offset: int = 0) -> str:
    """Decode and return uid string."""
    uid_length = message[offset]
    offset += 1
    uid = message[offset : uid_length + offset]

    return _encode_base5(uid + _uid_crc(uid))


def _encode_base5(data: bytes) -> str:
    """Encode bytes to base5 encided string."""
    key_string = "0123456789ABCDEFGHIJKLMNZPQRSTUV"
    number = int.from_bytes(data, "little")
    output: str = ""
    mask = (1 << 5) - 1
    while number:
        output = key_string[number & mask] + output
        number >>= 5

    return output


def _uid_crc(message: bytes) -> bytes:
    """Return UID crc."""
    crc_value = 0xA3A3
    for byte in message:
        crc_value = _uid_byte(crc_value ^ byte)

    return crc_value.to_bytes(byteorder="little", length=2)


def _uid_byte(byte: int) -> int:
    """Return CRC for a byte."""
    for _ in range(8):
        byte = (byte >> 1) ^ 0xA001 if byte & 1 else byte >> 1

    return byte



"""Contains mixers structure decoder."""
from __future__ import annotations

from typing import Final, List, Optional, Tuple

from pyplumio import util
from pyplumio.const import ATTR_MIXER_SENSORS
from pyplumio.helpers.typing import DeviceDataType
from pyplumio.structures import StructureDecoder, make_device_data

MIXER_TEMP: Final = "temp"
MIXER_TARGET_TEMP: Final = "target_temp"
MIXER_PUMP_OUTPUT: Final = "mixer_pump"
MIXER_DATA: Tuple[str, ...] = (
    MIXER_TEMP,
    MIXER_TARGET_TEMP,
    MIXER_PUMP_OUTPUT,
)


class MixersStructure(StructureDecoder):
    """Represents mixers data structure."""

    def decode(
        self, message: bytearray, offset: int = 0, data: Optional[DeviceDataType] = None
    ) -> Tuple[DeviceDataType, int]:
        """Decode bytes and return message data and offset."""
        mixers_number = message[offset]
        offset += 1
        mixer_sensors: List[DeviceDataType] = []
        for _ in range(mixers_number):
            mixer_sensor: DeviceDataType = {}
            mixer_sensor[MIXER_TEMP] = util.unpack_float(message[offset : offset + 4])[
                0
            ]
            mixer_sensor[MIXER_TARGET_TEMP] = message[offset + 4]
            mixer_outputs = message[offset + 6]
            mixer_sensor[MIXER_PUMP_OUTPUT] = bool(mixer_outputs & 0x01)
            mixer_sensors.append(mixer_sensor)
            offset += 8

        return make_device_data(data, {ATTR_MIXER_SENSORS: mixer_sensors}), offset



"""Contains network info structure decoder."""
from __future__ import annotations

from typing import Optional, Tuple

from pyplumio import util
from pyplumio.const import ATTR_NETWORK
from pyplumio.helpers.network_info import (
    EthernetParameters,
    NetworkInfo,
    WirelessParameters,
)
from pyplumio.helpers.typing import DeviceDataType
from pyplumio.structures import Structure, make_device_data


class NetworkInfoStructure(Structure):
    """Represents network info data structure."""

    def encode(self, data: DeviceDataType) -> bytearray:
        """Encode data and return bytearray."""
        message = bytearray()
        message += b"\x01"
        network_info = data[ATTR_NETWORK] if ATTR_NETWORK in data else NetworkInfo()
        message += util.ip4_to_bytes(network_info.eth.ip)
        message += util.ip4_to_bytes(network_info.eth.netmask)
        message += util.ip4_to_bytes(network_info.eth.gateway)
        message.append(network_info.eth.status)
        message += util.ip4_to_bytes(network_info.wlan.ip)
        message += util.ip4_to_bytes(network_info.wlan.netmask)
        message += util.ip4_to_bytes(network_info.wlan.gateway)
        message.append(network_info.server_status)
        message.append(network_info.wlan.encryption)
        message.append(network_info.wlan.signal_quality)
        message.append(network_info.wlan.status)
        message += b"\x00" * 4
        message.append(len(network_info.wlan.ssid))
        message += network_info.wlan.ssid.encode("utf-8")

        return message

    def decode(
        self, message: bytearray, offset: int = 0, data: Optional[DeviceDataType] = None
    ) -> Tuple[DeviceDataType, int]:
        """Decode bytes and return message data and offset."""
        network_info = NetworkInfo(
            eth=EthernetParameters(
                ip=util.ip4_from_bytes(message[offset : offset + 4]),
                netmask=util.ip4_from_bytes(message[offset + 4 : offset + 8]),
                gateway=util.ip4_from_bytes(message[offset + 8 : offset + 12]),
                status=bool(message[offset + 13]),
            ),
            wlan=WirelessParameters(
                ip=util.ip4_from_bytes(message[offset + 13 : offset + 17]),
                netmask=util.ip4_from_bytes(message[offset + 17 : offset + 21]),
                gateway=util.ip4_from_bytes(message[offset + 21 : offset + 25]),
                encryption=int(message[offset + 26]),
                signal_quality=int(message[offset + 27]),
                status=bool(message[offset + 28]),
                ssid=util.unpack_string(message, offset + 33),
            ),
            server_status=bool(message[offset + 25]),
        )

        return make_device_data(data, {ATTR_NETWORK: network_info}), offset



"""Contains constants."""
from __future__ import annotations

from typing import Final

STATE_ON: Final = "on"
STATE_OFF: Final = "off"

ADDR_BROADCAST: Final = 0
ADDR_ECONET: Final = 86

ATTR_NAME: Final = "name"
ATTR_VALUE: Final = "value"
ATTR_EXTRA: Final = "extra"
ATTR_BOILER_SENSORS: Final = "sensors"
ATTR_BOILER_PARAMETERS: Final = "parameters"
ATTR_PENDING_ALERTS: Final = "pending_alerts"
ATTR_ALERTS: Final = "alerts"
ATTR_FAN_POWER: Final = "fan_power"
ATTR_FRAME_VERSIONS: Final = "frames"
ATTR_FUEL_CONSUMPTION: Final = "fuel_consumption"
ATTR_FUEL_LEVEL: Final = "fuel_level"
ATTR_FUEL_BURNED: Final = "fuel_burned"
ATTR_LOAD: Final = "load"
ATTR_MIXERS: Final = "mixers"
ATTR_MIXER_SENSORS: Final = "mixer_sensors"
ATTR_MIXER_PARAMETERS: Final = "mixer_parameters"
ATTR_MODE: Final = "mode"
ATTR_MODULES: Final = "modules"
ATTR_POWER: Final = "power"
ATTR_THERMOSTAT: Final = "thermostat"
ATTR_THERMOSTATS: Final = "thermostats"
ATTR_TRANSMISSION: Final = "transmission"
ATTR_PRODUCT: Final = "product"
ATTR_NETWORK: Final = "network"
ATTR_VERSION: Final = "version"
ATTR_PASSWORD: Final = "password"
ATTR_SCHEMA: Final = "schema"
ATTR_REGDATA: Final = "regdata"
ATTR_LAMBDA_SENSOR: Final = "lambda"
ATTR_SCHEDULES: Final = "schedules"
ATTR_TYPE: Final = "type"
ATTR_SWITCH: Final = "switch"
ATTR_PARAMETER: Final = "parameter"
ATTR_SCHEDULE: Final = "schedule"



"""Contains message frames."""
from __future__ import annotations

from typing import ClassVar, Final

from pyplumio import util
from pyplumio.const import (
    ATTR_BOILER_SENSORS,
    ATTR_FAN_POWER,
    ATTR_FUEL_CONSUMPTION,
    ATTR_FUEL_LEVEL,
    ATTR_LOAD,
    ATTR_MODE,
    ATTR_POWER,
    ATTR_REGDATA,
    ATTR_THERMOSTAT,
    ATTR_TRANSMISSION,
)
from pyplumio.exceptions import VersionError
from pyplumio.frames import FrameTypes, Message
from pyplumio.helpers.typing import DeviceDataType, MessageType
from pyplumio.structures.frame_versions import FrameVersionsStructure
from pyplumio.structures.lambda_sensor import LambaSensorStructure
from pyplumio.structures.mixers import MixersStructure
from pyplumio.structures.modules import ModulesStructure
from pyplumio.structures.output_flags import OutputFlagsStructure
from pyplumio.structures.outputs import OutputsStructure
from pyplumio.structures.pending_alerts import PendingAlertsStructure
from pyplumio.structures.statuses import StatusesStructure
from pyplumio.structures.temperatures import TemperaturesStructure
from pyplumio.structures.thermostats import ThermostatsStructure

REGDATA_VERSION: Final = "1.0"


class RegulatorDataMessage(Message):
    """Represents current regulator data."""

    frame_type: ClassVar[int] = FrameTypes.MESSAGE_REGULATOR_DATA

    def decode_message(self, message: MessageType) -> DeviceDataType:
        """Decode frame message."""
        offset = 2
        frame_version = f"{message[offset+1]}.{message[offset]}"
        if frame_version != REGDATA_VERSION:
            raise VersionError(f"Unknown regdata version ({frame_version})")

        data, offset = FrameVersionsStructure(self).decode(message, offset + 2)
        data[ATTR_REGDATA] = message[offset:]

        return data


class SensorDataMessage(Message):
    """Represents current device state."""

    frame_type: ClassVar[int] = FrameTypes.MESSAGE_SENSOR_DATA

    def decode_message(self, message: MessageType) -> DeviceDataType:
        """Decode frame message."""
        sensors, offset = FrameVersionsStructure(self).decode(message, offset=0)
        sensors[ATTR_MODE] = message[offset]
        sensors, offset = OutputsStructure(self).decode(message, offset + 1, sensors)
        sensors, offset = OutputFlagsStructure(self).decode(message, offset, sensors)
        sensors, offset = TemperaturesStructure(self).decode(message, offset, sensors)
        sensors, offset = StatusesStructure(self).decode(message, offset, sensors)
        sensors, offset = PendingAlertsStructure(self).decode(message, offset, sensors)
        sensors[ATTR_FUEL_LEVEL] = message[offset]
        sensors[ATTR_TRANSMISSION] = message[offset + 1]
        sensors[ATTR_FAN_POWER] = util.unpack_float(message[offset + 2 : offset + 6])[0]
        sensors[ATTR_LOAD] = message[offset + 6]
        sensors[ATTR_POWER] = util.unpack_float(message[offset + 7 : offset + 11])[0]
        sensors[ATTR_FUEL_CONSUMPTION] = util.unpack_float(
            message[offset + 11 : offset + 15]
        )[0]
        sensors[ATTR_THERMOSTAT] = message[offset + 15]
        sensors, offset = ModulesStructure(self).decode(message, offset + 16, sensors)
        sensors, offset = LambaSensorStructure(self).decode(message, offset, sensors)
        sensors, offset = ThermostatsStructure(self).decode(message, offset, sensors)
        sensors, offset = MixersStructure(self).decode(message, offset, sensors)

        return {ATTR_BOILER_SENSORS: sensors}



"""Contains modules structure decoder."""
from __future__ import annotations

import struct
from typing import Final, Optional, Tuple

from pyplumio.const import ATTR_MODULES
from pyplumio.helpers.product_info import ConnectedModules
from pyplumio.helpers.typing import DeviceDataType
from pyplumio.structures import StructureDecoder, make_device_data

MODULE_A: Final = "module_a"
MODULE_B: Final = "module_b"
MODULE_C: Final = "module_c"
MODULE_LAMBDA: Final = "module_lambda"
MODULE_ECOSTER: Final = "module_ecoster"
MODULE_PANEL: Final = "module_panel"
MODULES: Tuple[str, ...] = (
    MODULE_A,
    MODULE_B,
    MODULE_C,
    MODULE_LAMBDA,
    MODULE_ECOSTER,
    MODULE_PANEL,
)


def _get_module_version(
    module_name: str, message: bytearray, offset: int = 0
) -> Tuple[Optional[str], int]:
    """Get module version from a message."""
    if message[offset] == 0xFF:
        return None, (offset + 1)

    version_data = struct.unpack("<BBB", message[offset : offset + 3])
    module_version = ".".join(str(i) for i in version_data)
    offset += 3

    if module_name == MODULE_A:
        vendor_code, vendor_version = struct.unpack("<BB", message[offset : offset + 2])
        module_version += f".{chr(vendor_code)}{str(vendor_version)}"
        offset += 2

    return module_version, offset


class ModulesStructure(StructureDecoder):
    """Represents modules data structure."""

    def decode(
        self, message: bytearray, offset: int = 0, data: Optional[DeviceDataType] = None
    ) -> Tuple[DeviceDataType, int]:
        """Decode bytes and return message data and offset."""
        connected_modules = ConnectedModules()
        for module_name in MODULES:
            module_version, offset = _get_module_version(module_name, message, offset)
            setattr(connected_modules, module_name, module_version)

        return make_device_data(data, {ATTR_MODULES: connected_modules}), offset



"""Contains data type representations."""
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Optional, Tuple, Type

from pyplumio import util
from pyplumio.helpers.typing import BytesType


class DataType(ABC):
    """Represents base data type."""

    _data: BytesType = bytearray()

    def __init__(self, data: Optional[BytesType] = None, size: Optional[int] = None):
        """Initialize new Data Type object."""
        if size is None:
            size = self.size

        if data is not None:
            self._data = data[0:size] if size > 0 else data

    def __eq__(self, other) -> bool:
        """Check if two data types are equal."""
        return self._data == other._data and self.size == other.size

    def __repr__(self) -> str:
        """Return serializable string representation of the class."""
        return f"{self.__class__.__name__}(data={self._data!r}, size={self.size})"

    def unpack(self, data: BytesType):
        """Unpack data to a given type."""
        self._data = data[0 : self.size]

    @property
    @abstractmethod
    def value(self):
        """Return data value."""

    @property
    @abstractmethod
    def size(self) -> int:
        """Return data length."""


class Undefined0(DataType):
    """Represents undefined zero-byte data type."""

    @property
    def value(self) -> None:
        """Return data value."""
        return None

    @property
    def size(self) -> int:
        """Return data length."""
        return 0


class SignedChar(DataType):
    """Represents signed char data type."""

    @property
    def value(self) -> int:
        """Return data value."""
        return util.unpack_char(self._data)[0]

    @property
    def size(self) -> int:
        """Return data length."""
        return 1


class Short(DataType):
    """Represents 16 bit integer data type."""

    @property
    def value(self) -> int:
        """Return data value."""
        return util.unpack_short(self._data)[0]

    @property
    def size(self) -> int:
        """Return data length."""
        return 2


class Int(DataType):
    """Represents 32 bit integer data type."""

    @property
    def value(self) -> int:
        """Return data value."""
        return util.unpack_int(self._data)[0]

    @property
    def size(self) -> int:
        """Return data length."""
        return 4


class Byte(DataType):
    """Represents byte data type."""

    @property
    def value(self) -> int:
        """Return data value."""
        return ord(self._data)

    @property
    def size(self) -> int:
        """Return data length."""
        return 1


class UnsignedShort(DataType):
    """Represents unsigned 16 bit integer data type."""

    @property
    def value(self) -> int:
        """Return data value."""
        return util.unpack_ushort(self._data)

    @property
    def size(self) -> int:
        """Return data length."""
        return 2


class UnsignedInt(DataType):
    """Represents unsigned 32 bit integer data type."""

    @property
    def value(self) -> int:
        """Return data value."""
        return util.unpack_uint(self._data)[0]

    @property
    def size(self) -> int:
        """Return data length."""
        return 4


class Float(DataType):
    """Representats float data type."""

    @property
    def value(self) -> float:
        """Return data value."""
        return util.unpack_float(self._data)[0]

    @property
    def size(self) -> int:
        """Return data length."""
        return 4


class Undefined8(DataType):
    """Represents undefined data type."""

    @property
    def value(self) -> None:
        """Return data value."""
        return None

    @property
    def size(self) -> int:
        """Return data length."""
        return 0


class Double(DataType):
    """Represents double data type."""

    @property
    def value(self) -> float:
        """Return data value."""
        return util.unpack_double(self._data)[0]

    @property
    def size(self) -> int:
        """Return data length."""
        return 8


class Boolean(DataType):
    """Represents boolean data type."""

    _index: int = 0

    def __init__(self, data: Optional[BytesType] = None, size: Optional[int] = None):
        """Initialize new Boolean object."""
        self._index = 0
        super().__init__(data, size=1)

    def index(self, index: int) -> int:
        """Return next bit index in the bit array."""
        self._index = index
        return 0 if self._index == 7 else self._index + 1

    def unpack(self, data: BytesType) -> None:
        """Unpack data to with given type."""
        self._data = data[0:1]

    @property
    def value(self) -> bool:
        """Return data value."""
        return bool(ord(self._data) & (1 << self._index))

    @property
    def size(self) -> int:
        """Return data length."""
        return 1 if self._index == 7 else 0


class Int64(DataType):
    """Represents 64 bit signed integer data type."""

    @property
    def value(self) -> int:
        """Return data value."""
        return util.unpack_int64(self._data)[0]

    @property
    def size(self) -> int:
        """Return data length."""
        return 8


class UInt64(DataType):
    """Represents 64 bit unsigned integer data type."""

    @property
    def value(self) -> int:
        """Return data value."""
        return util.unpack_uint64(self._data)[0]

    @property
    def size(self) -> int:
        """Return data length."""
        return 8


class IPv4(DataType):
    """Represents IPv4 address data type."""

    @property
    def value(self) -> str:
        """Return data value."""
        return util.ip4_from_bytes(self._data)

    @property
    def size(self) -> int:
        """Return data length."""
        return 4


class IPv6(DataType):
    """Represents IPv6 address data type."""

    @property
    def value(self) -> str:
        """Return data value."""
        return util.ip6_from_bytes(self._data)

    @property
    def size(self) -> int:
        """Return data length."""
        return 16


class String(DataType):
    """Represents string data type."""

    def __init__(self, data: Optional[BytesType] = None, size: Optional[int] = None):
        """Initialize new String object."""
        super().__init__(data, size=-1)

    def unpack(self, data: BytesType):
        """Unpack data to a given type."""
        self._data = data
        super().unpack(self._data)

    @property
    def value(self) -> str:
        """Return data value."""
        value = ""
        offset = 0
        if offset in self._data:
            while not self._data[offset] == 0:
                value += chr(self._data[offset])
                offset += 1

        return value

    @property
    def size(self) -> int:
        """Return data length."""
        return len(self.value) + 1


DATA_TYPES: Tuple[Type[DataType], ...] = (
    Undefined0,
    SignedChar,
    Short,
    Int,
    Byte,
    UnsignedShort,
    UnsignedInt,
    Float,
    Undefined8,
    Double,
    Boolean,
    String,
    String,
    Int64,
    UInt64,
    IPv4,
    IPv6,
)



"""Contains frame versions storage helper."""
from __future__ import annotations

from typing import TYPE_CHECKING, Final

from pyplumio.frames import get_frame_handler, is_known_frame_type
from pyplumio.helpers.factory import factory
from pyplumio.helpers.typing import VersionsInfoType

if TYPE_CHECKING:
    from pyplumio.devices import Device

DEFAULT_FRAME_VERSION: Final = 0


class FrameVersions:
    """Represents frame versions storage."""

    versions: VersionsInfoType
    device: Device

    def __init__(self, device: Device):
        """Initialize Frame Versions object."""
        self.versions = {}
        self.device = device

    async def async_update(self, frame_versions: VersionsInfoType) -> None:
        """Check versions and fetch outdated frames."""
        for frame_type, version in frame_versions.items():
            if is_known_frame_type(frame_type) and (
                frame_type not in self.versions or self.versions[frame_type] != version
            ):
                # We don't have this frame or it's version has changed.
                request = factory(
                    get_frame_handler(frame_type), recipient=self.device.address
                )
                self.device.queue.put_nowait(request)
                self.versions[frame_type] = version
